# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import deque

import numpy as np

from qililab.exceptions.variable_allocated import VariableAllocated
from qililab.qprogram.blocks import Average, Block, ForLoop, InfiniteLoop, Loop, Parallel
from qililab.qprogram.variable import Domain, FloatVariable, IntVariable, Variable
from qililab.yaml import yaml


@yaml.register_class
class VariableInfo:
    def __init__(self) -> None:
        self.is_allocated: bool = False
        self.allocated_by: Block | None = None

    def allocate(self, block: Block):
        self.is_allocated = True
        self.allocated_by = block

    def free(self):
        self.is_allocated = False
        self.allocated_by = None


@yaml.register_class
class StructuredProgram:
    """Represents a structured quantum program with various control flow blocks."""

    def __init__(self) -> None:
        """Initializes a StructuredProgram instance, setting up the body, block stack, variables, and buses."""
        self._body: Block = Block()
        self._block_stack: deque[Block] = deque([self._body])
        self._variables: dict[Variable, VariableInfo] = {}
        self._buses: set[str] = set()

    def _append_to_block_stack(self, block: Block):
        """Appends a block to the internal block stack.

        Args:
            block (Block): The block to append.
        """
        self._block_stack.append(block)

    def _pop_from_block_stack(self):
        """Removes and returns the last block from the block stack.

        Returns:
            Block: The popped block.
        """
        return self._block_stack.pop()

    @property
    def _active_block(self) -> Block:
        """Returns the currently active block on top of the block stack.

        Returns:
            Block: The active block.
        """
        return self._block_stack[-1]

    @property
    def body(self) -> Block:
        """Get the body of the QProgram

        Returns:
            Block: The block of the body
        """
        return self._body

    @property
    def buses(self) -> set[str]:
        """Get the buses.

        Returns:
            set[str]: A set of the names of the buses
        """
        return self._buses

    @property
    def variables(self) -> list[Variable]:
        """Get the variables.

        Returns:
            list[Variable]: A list of variables
        """
        return list(self._variables)

    def insert_block(self, block: Block):
        self._active_block.append(block)

    def block(self):
        """Define a generic block for scoping operations.

        Blocks need to open a scope.

        Returns:
            Block: The block.

        Examples:

            >>> with qp.block() as block:
            >>> # operations that shall be executed in the block
        """
        return StructuredProgram._BlockContext(program=self)

    def for_loop(self, variable: Variable, start: int | float, stop: int | float, step: int | float = 1):
        """Define a for_loop block to iterate values over a variable.

        Blocks need to open a scope.

        Args:
            variable (Variable): The variable to be affected from the loop.
            start (int | float): The start value.
            stop (int | float): The stop value.
            step (int | float, optional): The step value. Defaults to 1.

        Returns:
            Loop: The loop block.

        Examples:

            >>> variable = qp.variable(int)
            >>> with qp.for_loop(variable=variable, start=0, stop=100, step=5)):
            >>> # operations that shall be executed in the for_loop block
        """

        return StructuredProgram._ForLoopContext(program=self, variable=variable, start=start, stop=stop, step=step)

    def loop(self, variable: Variable, values: np.ndarray):
        """Define a loop block to iterate values over a variable.

        Blocks need to open a scope.

        Args:
            variable (Variable): The variable to be affected from the loop.
            values (np.ndarray): The values to iterate over.

        Returns:
            Loop: The loop block.

        Examples:

            >>> variable = qp.variable(int)
            >>> with qp.loop(variable=variable, values=np.array(range(100))):
            >>> # operations that shall be executed in the loop block
        """

        return StructuredProgram._LoopContext(program=self, variable=variable, values=values)

    def infinite_loop(self):
        """Define an infinite loop.

        Blocks need to open a scope.

        Examples:

            >>> with qp.infinite_loop():
            >>> # operations that shall be executed in the infinite loop block

        Returns:
            InfiniteLoop: The infinite loop block.
        """
        return StructuredProgram._InfiniteLoopContext(program=self)

    def parallel(self, loops: list[Loop | ForLoop]):
        """Define a block for running multiple loops in parallel.

        Blocks need to open a scope.

        Examples:
            >>> gain = qp.variable(float)
            >>> frequency = qp.variable(float)
            >>> with qp.parallel(loops=[ForLoop(variable=frequency, start=0, stop=100, step=10),
                                        ForLoop(variable=gain, start=0.0, stop=1.0, step=0.1)]):
            >>> # operations that shall be executed in the block

        Args:
            loops (list[Loop  |  ForLoop]): The loops to run in parallel

        Returns:
            Parallel: The parallel block.
        """
        return StructuredProgram._ParallelContext(program=self, loops=loops)

    def average(self, shots: int):
        """Define an acquire loop block with averaging in real time.

        Blocks need to open a scope.

        Args:
            iterations (int): The number of acquire iterations.

        Returns:
            Average: The average block.

        Examples:

            >>> with qp.average(shots=1000):
            >>> # operations that shall be executed in the average block
        """
        return StructuredProgram._AverageContext(program=self, shots=shots)

    def variable(self, label: str, domain: Domain, type: type[int | float] | None = None):
        """Declare a variable.

        Args:
            type (int | float): The type of the variable.

        Raises:
            NotImplementedError: If an unsupported type is provided.

        Returns:
            IntVariable | FloatVariable: The variable.
        """

        def _int_variable(label: str, domain: Domain = Domain.Scalar) -> IntVariable:
            variable = IntVariable(label, domain)
            self._variables[variable] = VariableInfo()
            return variable

        def _float_variable(label: str, domain: Domain = Domain.Scalar) -> FloatVariable:
            variable = FloatVariable(label, domain)
            self._variables[variable] = VariableInfo()
            return variable

        if domain is Domain.Scalar and type is None:
            raise ValueError("You must specify a type in a scalar variable.")
        if domain is not Domain.Scalar and type is not None:
            raise ValueError("When declaring a variable of a specific domain, its type is inferred by its domain.")

        if domain is Domain.Scalar:
            if type is int:
                return _int_variable(label, domain)
            if type is float:
                return _float_variable(label, domain)

        if domain == Domain.Time:
            return _int_variable(label, domain)
        if domain in [Domain.Frequency, Domain.Phase, Domain.Voltage, Domain.Flux]:
            return _float_variable(label, domain)
        raise NotImplementedError

    class _BlockContext:
        def __init__(self, program: "StructuredProgram"):
            self.program = program
            self.block: Block = Block()

        def __enter__(self):
            self.program._append_to_block_stack(block=self.block)
            return self.block

        def __exit__(self, exc_type, exc_value, exc_tb):
            block = self.program._pop_from_block_stack()
            self.program._active_block.append(block)

    class _ForLoopContext(_BlockContext):
        def __init__(
            self,
            program: "StructuredProgram",
            variable: Variable,
            start: int | float,
            stop: int | float,
            step: int | float,
        ):
            self.program = program
            self.block: ForLoop = ForLoop(variable=variable, start=start, stop=stop, step=step)

        def __enter__(self) -> ForLoop:
            if self.program._variables[self.block.variable].is_allocated:
                raise VariableAllocated(self.block.variable)
            self.program._variables[self.block.variable].allocate(self.block)
            return super().__enter__()

        def __exit__(self, exc_type, exc_value, exc_tb):
            self.program._variables[self.block.variable].free()
            super().__exit__(exc_type, exc_value, exc_tb)

    class _LoopContext(_BlockContext):
        def __init__(self, program: "StructuredProgram", variable: Variable, values: np.ndarray):
            self.program = program
            self.block: Loop = Loop(variable=variable, values=values)

        def __enter__(self) -> Loop:
            if self.program._variables[self.block.variable].is_allocated:
                raise VariableAllocated(self.block.variable)
            self.program._variables[self.block.variable].allocate(self.block)
            return super().__enter__()

        def __exit__(self, exc_type, exc_value, exc_tb):
            self.program._variables[self.block.variable].free()
            super().__exit__(exc_type, exc_value, exc_tb)

    class _InfiniteLoopContext(_BlockContext):
        def __init__(self, program: "StructuredProgram"):
            self.program = program
            self.block: InfiniteLoop = InfiniteLoop()

    class _ParallelContext(_BlockContext):
        def __init__(self, program: "StructuredProgram", loops: list[Loop | ForLoop]):
            self.program = program
            self.block: Parallel = Parallel(loops=loops)

        def __enter__(self) -> Parallel:
            for loop in self.block.loops:
                if self.program._variables[loop.variable].is_allocated:
                    raise VariableAllocated(loop.variable)
                self.program._variables[loop.variable].allocate(loop)
            return super().__enter__()

        def __exit__(self, exc_type, exc_value, exc_tb):
            for loop in self.block.loops:
                self.program._variables[loop.variable].free()
            super().__exit__(exc_type, exc_value, exc_tb)

    class _AverageContext(_BlockContext):
        def __init__(self, program: "StructuredProgram", shots: int):
            self.program = program
            self.block: Average = Average(shots=shots)



import os
import shutil
import time
from unittest.mock import MagicMock, patch

import pytest
import qililab as ql
from IPython.testing.globalipapp import start_ipython

slurm_job_data_test = "slurm_job_data_test"


@pytest.fixture(scope="session")
def session_ip():
    """Get an IPython shell"""
    yield start_ipython()


@pytest.fixture(scope="function")
def ip(session_ip):
    """Prepare IPython shell for running unit tests"""
    session_ip.run_cell(raw_cell="import qililab.slurm")
    yield session_ip


class TestSubmitJob:
    def teardown_method(self):
        """Teardown method to make sure all files are deleted."""
        if os.path.exists(slurm_job_data_test):
            shutil.rmtree(slurm_job_data_test)

    def test_submit_job(self, ip):
        ip.run_cell(raw_cell="a=1\nb=1")
        ip.run_cell_magic(
            magic_name="submit_job",
            line=f"-o results -g aQPU1 -l {slurm_job_data_test} -n unit_test -e local",
            cell="results = a+b ",
        )
        time.sleep(4)
        assert ip.user_global_ns["results"].result() == 2

    def test_submit_job_output_not_assigned(self, ip):
        """Check ValueError is raised in case values are not assigned to the output variable within the magic cell."""
        ip.run_cell(raw_cell="a=1\nb=1")
        with pytest.raises(
            ValueError, match="Output variable 'results' was not assigned to any value inside the cell!"
        ):
            ip.run_cell_magic(
                magic_name="submit_job",
                line=f"-o results -g aQPU1 -l {slurm_job_data_test} -n unit_test -e local",
                cell="a+b",
            )

    def test_submit_job_no_gres_provided(self, ip):
        """Check no Error is raised in case GRES is not provided."""
        ip.run_cell(raw_cell="a=1\nb=1")

        ip.run_cell_magic(
            magic_name="submit_job",
            line=f"-o results -l {slurm_job_data_test} -n unit_test -e local",
            cell="results = a+b ",
        )

    def test_submit_job_with_random_file_in_logs_folder(self, ip):
        """Check non-submitit files are deleted if found in logs folder"""
        ip.run_cell(
            raw_cell=f"import os\nfolder_path='{slurm_job_data_test}'\nfile_name='abc.py'\nfile_path = os.path.join(folder_path, file_name)\nif not os.path.exists(folder_path):\tos.makedirs(folder_path)\nopen(file_path, 'w')\na=1\nb=1"
        )
        ip.run_cell_magic(
            magic_name="submit_job",
            line=f"-o results -g aQPU1 -l {slurm_job_data_test} -n unit_test -e local",
            cell="results=a+b",
        )
        time.sleep(4)  # give time to ensure processes are finished
        assert os.path.isfile(os.path.join(slurm_job_data_test, "abc.py")) is False

    def test_submit_job_delete_info_from_past_jobs(self, ip):
        """Check only a certain amount files are kept in the logs folder"""
        ql.slurm.num_files_to_keep = 8
        ip.run_cell(raw_cell="a=1\nb=1")
        for _ in range(int(ql.slurm.num_files_to_keep / 4)):
            ip.run_cell_magic(
                magic_name="submit_job",
                line=f"-o results -g aQPU1 -l {slurm_job_data_test} -n unit_test -e local",
                cell="results=a+b",
            )
            time.sleep(2)  # give time submitit to create the files

        assert (
            len([f for f in os.listdir(slurm_job_data_test) if os.path.isfile(os.path.join(slurm_job_data_test, f))])
            == ql.slurm.num_files_to_keep
        )
        ip.run_cell_magic(
            magic_name="submit_job",
            line=f"-o results -g aQPU1 -l {slurm_job_data_test} -n unit_test -e local",
            cell="results=a+b",
        )
        time.sleep(2)
        assert (
            len([f for f in os.listdir(slurm_job_data_test) if os.path.isfile(os.path.join(slurm_job_data_test, f))])
            == ql.slurm.num_files_to_keep
        )
        assert ip.user_global_ns["results"].result() == 2

    def test_setting_parameters(self, ip):
        """Test that the parameters of the magic method work properly."""
        ip.run_cell(raw_cell="a=1\nb=1")
        with patch("qililab.slurm.AutoExecutor") as executor:
            executor_instance = MagicMock()
            executor.return_value = executor_instance
            with patch("qililab.slurm.os"):
                ip.run_cell_magic(
                    magic_name="submit_job",
                    line=f"-o results -p debug -l {slurm_job_data_test} -n unit_test -e local -t 5 --begin now+1hour -lp true -g aQPU1",
                    cell="results=a+b",
                )
        executor.assert_called_once_with(folder=slurm_job_data_test, cluster="local")
        executor_instance.update_parameters.assert_called_once_with(
            slurm_partition="debug",
            name="unit_test",
            timeout_min=5,
            slurm_additional_parameters={"begin": "now+1hour", "nice": 1000000, "gres":"aQPU1:1"},
        )



import numpy as np
import pytest

from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix, FluxVector


@pytest.fixture(name="crosstalk_array_buses")
def get_xtalk_array():
    """Get a tuple of crosstalk array and buses"""
    xtalk_array = np.array([[1, 2, 3], [0.1, 0.2, 0.3], [0, 1, 0]])
    buses = ["flux_0", "flux_1", "flux_2"]
    return (xtalk_array, buses)


@pytest.fixture(name="flux_vector_dict")
def get_flux_vector_dict():
    return {"flux_0": 0.5, "flux_1": 1, "flux_2": 0}


@pytest.fixture(name="flux_vector")
def get_flux_vector(flux_vector_dict):
    """Get a flux vector from CrosstalkMatrix.FluxVector"""
    return FluxVector.from_dict(flux_vector_dict)


@pytest.fixture(name="crosstalk_matrix")
def get_xtalk_matrix(crosstalk_array_buses):
    """Crosstalk matrix for the following xtalk
            flux_0  flux_1  flux_2
    flux_0  1       2       3
    flux_1  0.1     0.2     0.3
    flux_2  0       1       0
    """
    return CrosstalkMatrix.from_array(buses=crosstalk_array_buses[1], matrix_array=crosstalk_array_buses[0])


class TestFluxVector:
    """Unit test for the FluxVector class"""

    def test_init(self):
        flux_vector = FluxVector()
        assert flux_vector.bias_vector == {}

    def test_from_dict(self, flux_vector, flux_vector_dict):
        assert flux_vector.flux_vector == flux_vector_dict

    def test_to_dict(self, flux_vector, flux_vector_dict):
        assert flux_vector.to_dict() == flux_vector_dict

    def test_get(self, flux_vector):
        assert flux_vector["flux_0"] == 0.5

    def test_set_crosstalk_from_bias(self, flux_vector, crosstalk_matrix):
        bias_vector = flux_vector.set_crosstalk_from_bias(crosstalk_matrix)
        assert flux_vector.crosstalk == crosstalk_matrix
        assert flux_vector.bias_vector == bias_vector

    def test_to_dict_with_crosstalk(self, flux_vector, crosstalk_matrix):
        bias_vector = flux_vector.set_crosstalk_from_bias(crosstalk_matrix)
        assert flux_vector.to_dict() == bias_vector

    def test_get_with_crosstalk(self, flux_vector, crosstalk_matrix):
        bias_vector = flux_vector.set_crosstalk_from_bias(crosstalk_matrix)
        assert flux_vector["flux_0"] == bias_vector["flux_0"]


class TestCrosstalkMatrix:
    """Unit tests checking the Calibration attributes and methods."""

    def test_from_array(self, crosstalk_matrix, crosstalk_array_buses):
        """Test that from array works as intended"""
        xtalk_array, buses = crosstalk_array_buses
        assert (
            sum(
                crosstalk_matrix[bus1][bus2] == xtalk_array[i, j]
                for i, bus1 in enumerate(buses)
                for j, bus2 in enumerate(buses)
            )
            == xtalk_array.shape[0] * xtalk_array.shape[1]
        )

    def test_to_array(self, crosstalk_matrix, crosstalk_array_buses):
        assert np.allclose(crosstalk_matrix.to_array(), crosstalk_array_buses[0])

    def test_inverse(self, crosstalk_matrix):
        inv_array = np.linalg.inv(crosstalk_matrix.to_array())
        inv_matrix = crosstalk_matrix.inverse()
        assert (
            sum(
                inv_matrix[bus1][bus2] == inv_array[i, j]
                for i, bus1 in enumerate(inv_matrix.matrix.keys())
                for j, bus2 in enumerate(inv_matrix.matrix.keys())
            )
            == inv_array.shape[0] * inv_array.shape[1]
        )

    def test_init(self):
        """Test init method"""
        crosstalk_matrix = CrosstalkMatrix()

        assert isinstance(crosstalk_matrix, CrosstalkMatrix)
        assert isinstance(crosstalk_matrix.matrix, dict)
        assert len(crosstalk_matrix.matrix) == 0

    def test_set_get_methods(self):
        """Test __setitem__ and __getitem__ methods"""
        crosstalk_matrix = CrosstalkMatrix()
        crosstalk_matrix["bus1"]["bus2"] = 0.5

        assert isinstance(crosstalk_matrix["bus1"], dict)
        assert len(crosstalk_matrix["bus1"]) == 1
        assert crosstalk_matrix["bus1"]["bus2"] == 0.5

        # Reshape a given key
        crosstalk_matrix["bus1"] = {"bus2": 0.6}
        assert isinstance(crosstalk_matrix["bus1"], dict)
        assert len(crosstalk_matrix["bus1"]) == 1
        assert crosstalk_matrix["bus1"]["bus2"] == 0.6

        # New key
        crosstalk_matrix["bus2"] = {"bus1": 0.1}
        assert isinstance(crosstalk_matrix["bus2"], dict)
        assert len(crosstalk_matrix["bus2"]) == 1
        assert crosstalk_matrix["bus2"]["bus1"] == 0.1

    def test_set_offset(self):
        """Test set_offset function"""
        crosstalk_matrix = CrosstalkMatrix()
        crosstalk_matrix["bus1"]["bus1"] = 1
        crosstalk_matrix["bus1"]["bus2"] = 0.5
        crosstalk_matrix["bus2"]["bus1"] = 0.7
        crosstalk_matrix["bus2"]["bus2"] = 1

        crosstalk_matrix.set_offset(offset={"bus1": -1.0, "bus2": 0.5})

        assert isinstance(crosstalk_matrix.flux_offsets, dict)
        assert crosstalk_matrix.flux_offsets == {"bus1": -1.0, "bus2": 0.5}

    def test_str_method(self):
        """Test __str__ method"""
        crosstalk_matrix = CrosstalkMatrix()
        crosstalk_matrix["bus1"]["bus1"] = -0.5
        crosstalk_matrix["bus1"]["bus2"] = 0.5
        crosstalk_matrix["bus2"]["bus1"] = 0.7
        crosstalk_matrix["bus2"]["bus2"] = -0.5

        string = str(crosstalk_matrix)
        assert (
            string
            == """            bus1     bus2
bus1         -0.5      0.5
bus2          0.7     -0.5"""
        )

    def test_repr_method(self):
        """Test __repr__ method"""
        crosstalk_matrix = CrosstalkMatrix()
        crosstalk_matrix["bus1"]["bus1"] = -0.5
        crosstalk_matrix["bus1"]["bus2"] = 0.5
        crosstalk_matrix["bus2"]["bus1"] = 0.7
        crosstalk_matrix["bus2"]["bus2"] = -0.5

        representation = repr(crosstalk_matrix)
        assert (
            representation
            == "CrosstalkMatrix({'bus1': {'bus1': -0.5, 'bus2': 0.5}, 'bus2': {'bus1': 0.7, 'bus2': -0.5}})"
        )

    def test_from_buses_method(self):
        """Test from_buses class method"""
        buses = {
            "bus1": {"bus1": 1.0, "bus2": 1.0, "bus3": 1.0},
            "bus2": {"bus1": 1.0, "bus2": 1.0, "bus3": 1.0},
            "bus3": {"bus1": 1.0, "bus2": 1.0, "bus3": 1.0},
        }
        crosstalk_matrix = CrosstalkMatrix.from_buses(buses)

        assert isinstance(crosstalk_matrix, CrosstalkMatrix)
        assert isinstance(crosstalk_matrix.matrix, dict)

        assert "bus1" in crosstalk_matrix.matrix
        assert "bus2" in crosstalk_matrix.matrix["bus1"]
        assert "bus3" in crosstalk_matrix.matrix["bus1"]
        assert crosstalk_matrix["bus1"]["bus2"] == 1.0
        assert crosstalk_matrix["bus1"]["bus3"] == 1.0

        assert "bus2" in crosstalk_matrix.matrix
        assert "bus1" in crosstalk_matrix.matrix["bus2"]
        assert "bus3" in crosstalk_matrix.matrix["bus2"]
        assert crosstalk_matrix["bus2"]["bus1"] == 1.0
        assert crosstalk_matrix["bus2"]["bus3"] == 1.0

        assert "bus3" in crosstalk_matrix.matrix
        assert "bus1" in crosstalk_matrix.matrix["bus3"]
        assert "bus2" in crosstalk_matrix.matrix["bus3"]
        assert crosstalk_matrix["bus3"]["bus1"] == 1.0
        assert crosstalk_matrix["bus3"]["bus2"] == 1.0



""" Test Results """

import os
from warnings import catch_warnings

import numpy as np
import pytest

from qililab.result.qprogram.quantum_machines_measurement_result import QuantumMachinesMeasurementResult
from qililab.utils.serialization import deserialize, deserialize_from, serialize, serialize_to


@pytest.fixture(name="quantum_machines_measurement_result")
def fixture_quantum_machines_measurement_result():
    """fixture_quantum_machines_measurement_result

    Returns:
        quantum_machines_measurement_result: QuantumMachinesResult
    """
    return QuantumMachinesMeasurementResult(bus="readout", I=np.zeros(10), Q=np.zeros(10))


class TestsQMResult:
    """Test `QuantumMachinesResult` functionalities."""

    def test_init(self, quantum_machines_measurement_result: QuantumMachinesMeasurementResult):
        """Test the instantiation of QuantumMachinesMeasurementResult."""
        assert isinstance(quantum_machines_measurement_result, QuantumMachinesMeasurementResult)
        assert np.allclose(quantum_machines_measurement_result.I, np.zeros(10))
        assert np.allclose(quantum_machines_measurement_result.Q, np.zeros(10))

    def test_set_classification_threshold(self, quantum_machines_measurement_result: QuantumMachinesMeasurementResult):
        """Test the set_classification_threshold method."""
        quantum_machines_measurement_result.set_classification_threshold(0.3)
        assert quantum_machines_measurement_result._classification_threshold == 0.3

    def test_array(self, quantum_machines_measurement_result: QuantumMachinesMeasurementResult):
        """Test the array property returns the correct data."""
        assert isinstance(quantum_machines_measurement_result.array, np.ndarray)
        assert quantum_machines_measurement_result.array.shape == (2, 10)

    def test_serialization_method(self, quantum_machines_measurement_result: QuantumMachinesMeasurementResult):
        """Test serialization and deserialization works."""
        serialized = serialize(quantum_machines_measurement_result)
        deserialized_quantum_machines_measurement_result = deserialize(serialized, QuantumMachinesMeasurementResult)

        assert isinstance(deserialized_quantum_machines_measurement_result, QuantumMachinesMeasurementResult)

        serialize_to(quantum_machines_measurement_result, file="quantum_machines_measurement_result.yml")
        deserialized_quantum_machines_measurement_result = deserialize_from(
            "quantum_machines_measurement_result.yml", QuantumMachinesMeasurementResult
        )

        assert isinstance(deserialized_quantum_machines_measurement_result, QuantumMachinesMeasurementResult)

        os.remove("quantum_machines_measurement_result.yml")

    @pytest.mark.parametrize(
        "threshold, I, expected",
        [
            (0.0, np.ones(10), np.ones(10)),
            (0.5, np.zeros(10), np.zeros(10)),
            (0.5, np.array([0.2, 0.4, 0.6, 0.5, 0.9]), np.array([0.0, 0.0, 1.0, 1.0, 1.0])),
            (None, np.array([0.2, 0.4, 0.6, 0.5, 0.9]), np.array([0.0, 0.0, 0.0, 0.0, 0.0])),
        ],
    )
    def test_threshold_property(
        self, threshold, I, expected, quantum_machines_measurement_result: QuantumMachinesMeasurementResult
    ):
        """Test the `threshold` property."""
        quantum_machines_measurement_result._classification_threshold = threshold
        quantum_machines_measurement_result.I = I
        if threshold is not None:
            np.testing.assert_equal(quantum_machines_measurement_result.threshold, expected)

        else:
            with catch_warnings(record=True) as w:
                np.testing.assert_equal(quantum_machines_measurement_result.threshold, expected)
                assert (
                    w[0].message.args[0] == "Classification threshold is not specified, returning a `np.zeros` array."  # type: ignore
                )



"""Tests for the SGS100A class."""

import re
import warnings
from unittest.mock import MagicMock, patch

import pytest

from qililab.instruments import SGS100A, ParameterNotFound
from qililab.typings.enums import Parameter


@pytest.fixture(name="sdg100a")
def fixture_sdg100a() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 80e6,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": True,
            "alc": True,
        }
    )
    sdg100a.device = MagicMock()
    return sdg100a


@pytest.fixture(name="sdg100a_iq")
def fixture_sdg100a_iq() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_iq = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 80e6,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": True,
            "alc": True,
        }
    )
    sdg100a_iq.device = MagicMock()
    return sdg100a_iq


@pytest.fixture(name="sdg100a_rf_off")
def fixture_sdg100a_rf_off() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_rf_off = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 80e6,
            "rf_on": False,
            "iq_modulation": True,
            "iq_wideband": True,
            "alc": True,
        }
    )
    sdg100a_rf_off.device = MagicMock()
    return sdg100a_rf_off


@pytest.fixture(name="sdg100a_alc_off")
def fixture_sdg100a_alc_off() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_alc_off = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 80e6,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": True,
            "alc": False,
        }
    )
    sdg100a_alc_off.device = MagicMock()
    return sdg100a_alc_off


@pytest.fixture(name="sdg100a_iq_mod_off")
def fixture_sdg100a_iq_mod_off() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_iq_mod_off = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 80e6,
            "rf_on": True,
            "iq_modulation": False,
            "iq_wideband": True,
            "alc": True,
        }
    )
    sdg100a_iq_mod_off.device = MagicMock()
    return sdg100a_iq_mod_off


@pytest.fixture(name="sdg100a_wideband_off")
def fixture_sdg100a_wideband_off() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_wideband_off = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 80e6,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": False,
            "alc": True,
        }
    )
    sdg100a_wideband_off.device = MagicMock()
    return sdg100a_wideband_off


@pytest.fixture(name="sdg100a_wideband_off_large_freq")
def fixture_sdg100a_wideband_off_large_freq() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_wideband_off_large_freq = SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 5e9,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": False,
            "alc": True,
        }
    )
    sdg100a_wideband_off_large_freq.device = MagicMock()
    return sdg100a_wideband_off_large_freq


@pytest.fixture(name="sdg100a_wrong_power")
def fixture_sdg100a_wrong_power() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_wrong_power = SGS100A(
        {
            "alias": "qdac",
            "power": 100,
            "frequency": 80e6,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": True,
            "alc": True,
        }
    )
    sdg100a_wrong_power.device = MagicMock()
    return sdg100a_wrong_power


@pytest.fixture(name="sdg100a_wrong_freq")
def fixture_sdg100a_wrong_freq() -> SGS100A:
    """Fixture that returns an instance of a dummy QDAC-II."""
    sdg100a_wrong_freq= SGS100A(
        {
            "alias": "qdac",
            "power": 10,
            "frequency": 13e9,
            "rf_on": True,
            "iq_modulation": True,
            "iq_wideband": True,
            "alc": True,
        }
    )
    sdg100a_wrong_freq.device = MagicMock()
    return sdg100a_wrong_freq


class TestSGS100A:
    """Unit tests checking the SGS100A attributes and methods"""

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.POWER, 0.01),
            (Parameter.LO_FREQUENCY, 6.0e09),
            (Parameter.RF_ON, True),
            (Parameter.RF_ON, False),
            (Parameter.IQ_MODULATION, True),
            (Parameter.IQ_MODULATION, False),
            (Parameter.ALC, True),
            (Parameter.ALC, False),
            (Parameter.IQ_WIDEBAND, True),
            (Parameter.IQ_WIDEBAND, False),
        ],
    )
    def test_set_parameter_method(
        self,
        sdg100a: SGS100A,
        parameter: Parameter,
        value: float,
    ):
        """Test setup method"""
        sdg100a.set_parameter(parameter=parameter, value=value)
        if parameter == Parameter.POWER:
            assert sdg100a.settings.power == value
        if parameter == Parameter.LO_FREQUENCY:
            assert sdg100a.settings.frequency == value
        if parameter == Parameter.RF_ON:
            assert sdg100a.settings.rf_on == value
        if parameter == Parameter.IQ_MODULATION:
            assert sdg100a.settings.iq_modulation == value

    def test_set_parameter_method_raises_error(self, sdg100a: SGS100A):
        """Test setup method"""
        with pytest.raises(ParameterNotFound):
            sdg100a.set_parameter(parameter=Parameter.BUS_FREQUENCY, value=123)

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            (Parameter.POWER, 10),
            (Parameter.LO_FREQUENCY, 80e6),
            (Parameter.RF_ON, True),
            (Parameter.IQ_MODULATION, True),
            (Parameter.ALC, True),
            (Parameter.IQ_WIDEBAND, True),
        ],
    )
    def test_get_parameter_method(
        self,
        sdg100a: SGS100A,
        parameter: Parameter,
        expected_value: float,
    ):
        """Test get_parameter method"""
        value = sdg100a.get_parameter(parameter=parameter)
        assert value == expected_value

    def test_get_parameter_method_raises_error(self, sdg100a: SGS100A):
        """Test get_parameter method"""
        with pytest.raises(ParameterNotFound):
            sdg100a.get_parameter(parameter=Parameter.BUS_FREQUENCY)

    @patch("warnings.warn")
    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_initial_setup_method(self, mock_get_rs_options, mock_warn, sdg100a: SGS100A):
        """Test initial setup method"""
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        sdg100a.initial_setup()
        sdg100a.device.power.assert_called_with(sdg100a.power)
        sdg100a.device.frequency.assert_called_with(sdg100a.frequency)
        sdg100a.device.on.assert_called_once()

        sdg100a.settings.rf_on = False
        sdg100a.initial_setup()
        sdg100a.device.power.assert_called_with(sdg100a.power)
        sdg100a.device.frequency.assert_called_with(sdg100a.frequency)
        sdg100a.device.off.assert_called_once()

        mock_warn.assert_any_call(
            "LO frequency below 2.5GHz only allows for IF sweeps of ±1.60e+07 Hz", ResourceWarning
        )

    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_initial_setup_method_iq_mod_off(self, mock_get_rs_options, sdg100a_iq_mod_off: SGS100A):
        """Test initial method when the runcard sets rf_on as False"""
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        sdg100a_iq_mod_off.initial_setup()
        assert sdg100a_iq_mod_off.settings.iq_modulation is False
        assert sdg100a_iq_mod_off.iq_modulation is False

    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_initial_setup_method_alc_off(self, mock_get_rs_options, sdg100a_alc_off: SGS100A):
        """Test initial method when the runcard sets rf_on as False"""
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        sdg100a_alc_off.initial_setup()
        assert sdg100a_alc_off.settings.alc is False
        assert sdg100a_alc_off.alc is False

    @patch("warnings.warn")
    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_initial_setup_method_wideband_off(self, mock_get_rs_options, mock_warn, sdg100a_wideband_off: SGS100A):
        """Test initial method when the runcard sets rf_on as False"""
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        sdg100a_wideband_off.initial_setup()
        assert sdg100a_wideband_off.settings.iq_wideband is False
        assert sdg100a_wideband_off.iq_wideband is False
        assert sdg100a_wideband_off.iq_modulation is True

        mock_warn.assert_any_call(
            "Deactivated wideband & LO frequency below 1GHz allows for IF sweeps of ±4.00e+06 Hz", ResourceWarning
        )

    @patch("warnings.warn")
    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_initial_setup_method_wideband_off_large_freq(
        self, mock_get_rs_options, mock_warn, sdg100a_wideband_off_large_freq: SGS100A
    ):
        """Test initial method when the runcard sets rf_on as False"""
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        sdg100a_wideband_off_large_freq.initial_setup()
        assert sdg100a_wideband_off_large_freq.settings.iq_wideband is False
        assert sdg100a_wideband_off_large_freq.iq_wideband is False
        assert sdg100a_wideband_off_large_freq.iq_modulation is True

        mock_warn.assert_any_call("Deactivated wideband allows for IF sweeps of ±50e6 Hz", ResourceWarning)

    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_initial_setup_sets_correct_freq_range(self, mock_get_rs_options, sdg100a: SGS100A):
        """Test initial setup method"""
        mock_get_rs_options.return_value = "Some,other,SGS-B112"
        sdg100a.initial_setup()
        assert sdg100a.freq_top_limit == 12.75e9
        assert sdg100a.freq_bot_limit == 1e6

        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        sdg100a.initial_setup()
        assert sdg100a.freq_top_limit == 12.75e9
        assert sdg100a.freq_bot_limit == 80e6

        mock_get_rs_options.return_value = "Some,other,SGS-B106"
        sdg100a.initial_setup()
        assert sdg100a.freq_top_limit == 6e9
        assert sdg100a.freq_bot_limit == 1e6

        mock_get_rs_options.return_value = "Some,other,SGS-B106V"
        sdg100a.initial_setup()
        assert sdg100a.freq_top_limit == 6e9
        assert sdg100a.freq_bot_limit == 80e6

    def test_get_rs_options(self, sdg100a: SGS100A):
        sdg100a.get_rs_options()
        sdg100a.device.ask.assert_called_once()

    def test_turn_on_method_rf_off(self, sdg100a_rf_off: SGS100A):
        """Test initial method when the runcard sets rf_on as False"""
        sdg100a_rf_off.turn_on()
        assert sdg100a_rf_off.settings.rf_on is False
        sdg100a_rf_off.device.off.assert_called_once()

    def test_turn_on_method(self, sdg100a: SGS100A):
        """Test turn_on method"""
        sdg100a.turn_on()
        assert sdg100a.settings.rf_on is True
        sdg100a.device.on.assert_called_once()

    def test_turn_off_method(self, sdg100a: SGS100A):
        """Test turn_off method"""
        sdg100a.turn_off()
        assert sdg100a.settings.rf_on is False
        sdg100a.device.off.assert_called_once()

    def test_reset_method(self, sdg100a: SGS100A):
        """Test reset method"""
        sdg100a.reset()

    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_raise_error_power_out_of_range(self, mock_get_rs_options, sdg100a_wrong_power: SGS100A):
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        error_string = "Value set for power is outside of the allowed range [-120,25]: 100"
        with pytest.raises(ValueError, match=re.escape(error_string)):
            sdg100a_wrong_power.initial_setup()

    @patch("qililab.instruments.rohde_schwarz.SGS100A.get_rs_options")
    def test_raise_error_freq_out_of_range(self, mock_get_rs_options, sdg100a_wrong_freq: SGS100A):
        mock_get_rs_options.return_value = "Some,other,SGS-B112V"
        error_string = "Value set for frequency is outside of the allowed range [80000000.0, 12750000000.0]: 13000000000.0"
        with pytest.raises(ValueError, match=re.escape(error_string)):
            sdg100a_wrong_freq.initial_setup()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrument class"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import get_type_hints

from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.platform.components.bus_element import BusElement
from qililab.settings import Settings
from qililab.typings import ChannelID, Device, InstrumentName, Parameter, ParameterValue


class Instrument(BusElement, ABC):
    """Abstract base class declaring the necessary attributes
    and methods for the instruments.

    Args:
        settings (Settings): Class containing the settings of the instrument.
    """

    name: InstrumentName

    @dataclass
    class InstrumentSettings(Settings):
        """Contains the settings of an instrument.

        Args:
            alias (str): Alias of the instrument.
        """

        alias: str

    settings: InstrumentSettings  # a subtype of settings must be specified by the subclass
    device: Device

    def is_device_active(self) -> bool:
        """Check wether or not the device is currently active, for instrument childs.

        Contrary to ``CheckDeviceInitialized``, we also check if the device is not None, since a ``disconnect()`` after
        initialization would set it to None.

        Returns:
            bool: Wether or not the device has been initialized.
        """
        return hasattr(self, "device") and self.device is not None

    def __init__(self, settings: dict):
        settings_class: type[Instrument.InstrumentSettings] = get_type_hints(self).get("settings")  # type: ignore
        self.settings = settings_class(**settings)

    def __str__(self):
        """String representation of an instrument."""
        return f"{self.alias}"

    @property
    def alias(self):
        """Instrument 'alias' property.

        Returns:
            str: settings.alias.
        """
        return self.settings.alias

    def is_awg(self) -> bool:
        """Returns True if instrument is an AWG."""
        return False

    def is_adc(self) -> bool:
        """Returns True if instrument is an ADC."""
        return False

    @check_device_initialized
    @abstractmethod
    def turn_on(self):
        """Turn on an instrument."""

    @check_device_initialized
    @abstractmethod
    def turn_off(self):
        """Turn off an instrument."""

    @check_device_initialized
    @abstractmethod
    def reset(self):
        """Reset instrument settings."""

    @check_device_initialized
    @abstractmethod
    def initial_setup(self):
        """Set initial instrument settings."""

    @abstractmethod
    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None) -> ParameterValue:
        """Gets the parameter of a specific instrument.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int, optional): Instrument channel, if multiple. Defaults to None.

        Returns:
            str | int | float | bool: Parameter value.
        """

    @log_set_parameter
    @abstractmethod
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Sets the parameter of a specific instrument.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (float | str | bool): value to update
            channel_id (int, optional): instrument channel to update, if multiple. Defaults to None.

        Returns:
            bool: True if the parameter is set correctly, False otherwise
        """


class ParameterNotFound(Exception):
    """Error raised when a parameter in an instrument is not found."""

    def __init__(
        self,
        instrument: Instrument,
        parameter: Parameter,
    ):
        self.message: str = (
            f"Could not find parameter {parameter} in instrument {instrument.name} with alias {instrument.alias}."
        )
        super().__init__(self.message)

    def __str__(self):
        return f"ParameterNotFound: {self.message}"



"""This file tests the the ``InstrumentController`` class"""

import io

import pytest
from ruamel.yaml import YAML

from qililab.constants import CONNECTION, INSTRUMENTCONTROLLER, RUNCARD
from qililab.instrument_controllers.rohde_schwarz import SGS100AController
from qililab.platform import Platform
from qililab.typings.enums import ConnectionName, InstrumentControllerName, Parameter
from tests.data import Galadriel
from tests.test_utils import build_platform


@pytest.fixture(name="platform")
def fixture_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=Galadriel.runcard)


@pytest.fixture(name="rs_settings")
def fixture_rs_settings():
    """Fixture that returns an instance of a dummy RS."""
    return {
        RUNCARD.NAME: InstrumentControllerName.ROHDE_SCHWARZ,
        RUNCARD.ALIAS: "rohde_schwarz_controller_0",
        Parameter.REFERENCE_CLOCK.value: "EXT",
        INSTRUMENTCONTROLLER.CONNECTION: {
            RUNCARD.NAME: ConnectionName.TCP_IP.value,
            CONNECTION.ADDRESS: "192.168.0.10",
        },
        INSTRUMENTCONTROLLER.MODULES: [
            {
                "alias": "rs_0",
                "slot_id": 0,
            }
        ],
    }


class TestConnection:
    """This class contains the unit tests for the ``InstrumentController`` class."""

    def test_error_raises_when_no_modules(self, platform: Platform, rs_settings):
        """Test that ensures an error raises when there is no module specifyed

        Args:
            platform (Platform): Platform
        """
        rs_settings[INSTRUMENTCONTROLLER.MODULES] = []
        name = rs_settings.pop(RUNCARD.NAME)
        with pytest.raises(ValueError, match=f"The {name.value} Instrument Controller requires at least ONE module."):
            SGS100AController(settings=rs_settings, loaded_instruments=platform.instruments)

    def test_error_raises_when_no_compatible_number_modules(self, platform: Platform, rs_settings):
        """Error raises when the number of modules of the instument controller
        is not compatible with the modules loded.

        Args:
            platform (Platform): Platform
        """
        rs_settings[INSTRUMENTCONTROLLER.MODULES] = [
            {
                "alias": "rs_0",
                "slot_id": 0,
            },
            {
                "alias": "rs_1",
                "slot_id": 1,
            },
        ]
        name = rs_settings.pop(RUNCARD.NAME)
        with pytest.raises(
            ValueError,
            match=f"The {name.value} Instrument Controller only supports 1 module/s.You have loaded 2 modules.",
        ):
            SGS100AController(settings=rs_settings, loaded_instruments=platform.instruments)

    def test_print_instrument_controllers(self, platform: Platform):
        """Test print instruments."""
        instr_cont = platform.instrument_controllers
        assert str(instr_cont) == str(YAML().dump(instr_cont.to_dict(), io.BytesIO()))

    def test_reset_to_dict(self, platform: Platform):
        """Test that the reset attribute gets reflected when calling the controller to_dict method."""
        instr_cont = platform.instrument_controllers
        controllers_dict = instr_cont.to_dict()

    def test_set_get_reset(self, platform: Platform):
        assert platform.get_parameter(alias="rohde_schwarz_controller_0", parameter=Parameter.RESET) == True
        platform.set_parameter(alias="rohde_schwarz_controller_0", parameter=Parameter.RESET, value=False)
        assert platform.get_parameter(alias="rohde_schwarz_controller_0", parameter=Parameter.RESET) == False

        with pytest.raises(ValueError):
            _ = platform.get_parameter(alias="rohde_schwarz_controller_0", parameter=Parameter.BUS_FREQUENCY)

        with pytest.raises(ValueError):
            _ = platform.set_parameter(alias="rohde_schwarz_controller_0", parameter=Parameter.BUS_FREQUENCY, value=1e9)



from unittest.mock import patch

import numpy as np
import pytest
import qpysequence as QPy

from qililab import Calibration, Domain, Gaussian, IQPair, QbloxCompiler, QProgram, Square
from qililab.qprogram.blocks import ForLoop
from tests.test_utils import is_q1asm_equal
from qililab.config import logger
import logging


def setup_q1asm(marker: str):
    return f"""
        setup:
                wait_sync        4
                set_mrk          {marker}
                upd_param        4
    """


@pytest.fixture(name="calibration")
def fixture_calibration() -> Calibration:
    calibration = Calibration()
    calibration.add_waveform(bus="drive_q0", name="Xpi", waveform=Square(1.0, 100))
    calibration.add_waveform(bus="drive_q1", name="Xpi", waveform=Square(1.0, 150))
    calibration.add_waveform(bus="drive_q2", name="Xpi", waveform=Square(1.0, 200))

    return calibration


@pytest.fixture(name="play_named_operation")
def fixture_play_named_operation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    qp = QProgram()
    qp.play(bus="drive", waveform="Xpi")
    qp.play(bus="drive", waveform=drag_wf)

    return qp

@pytest.fixture(name="measurement_blocked_operation")
def fixture_measurement_blocked_operation() -> QProgram:
    drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    with qp.block():
        qp.play(bus="drive", waveform=drag_wf)
        qp.measure(bus="readout", waveform=readout_pair, weights=weights_pair)

    return qp


@pytest.fixture(name="no_loops_all_operations")
def fixture_no_loops_all_operations() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    qp.set_frequency(bus="drive", frequency=300)
    qp.set_phase(bus="drive", phase=np.pi / 2)
    qp.reset_phase(bus="drive")
    qp.set_gain(bus="drive", gain=0.5)
    qp.set_offset(bus="drive", offset_path0=0.5, offset_path1=0.5)
    qp.play(bus="drive", waveform=drag_pair)
    qp.sync()
    qp.wait(bus="readout", duration=100)
    qp.play(bus="readout", waveform=readout_pair)
    qp.qblox.set_markers(bus="readout", mask="0111")
    qp.qblox.play(bus="readout", waveform=readout_pair, wait_time=4)
    qp.qblox.acquire(bus="readout", weights=weights_pair)
    return qp


@pytest.fixture(name="offset_no_path1")
def fixture_offset_no_path1() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp = QProgram()
    qp.set_offset(bus="drive", offset_path0=0.5)
    qp.play(bus="drive", waveform=drag_pair)
    return qp


@pytest.fixture(name="dynamic_wait")
def fixture_dynamic_wait() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp = QProgram()
    duration = qp.variable(label="time", domain=Domain.Time)
    with qp.for_loop(variable=duration, start=100, stop=200, step=10):
        qp.play(bus="drive", waveform=drag_pair)
        qp.wait(bus="drive", duration=duration)
    return qp


@pytest.fixture(name="dynamic_wait_multiple_buses")
def fixture_dynamic_wait_multiple_buses() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp = QProgram()
    duration = qp.variable(label="time", domain=Domain.Time)
    with qp.for_loop(variable=duration, start=100, stop=200, step=10):
        qp.play(bus="drive", waveform=drag_pair)
        qp.wait(bus="readout", duration=duration)
        qp.play(bus="readout", waveform=drag_pair)
    return qp


@pytest.fixture(name="dynamic_wait_multiple_buses_with_disable_autosync")
def fixture_dynamic_wait_multiple_buses_with_disable_autosync() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp = QProgram()
    qp.qblox.disable_autosync = True
    duration = qp.variable(label="time", domain=Domain.Time)
    with qp.for_loop(variable=duration, start=100, stop=200, step=10):
        qp.play(bus="drive", waveform=drag_pair)
        qp.wait(bus="readout", duration=duration)
        qp.play(bus="readout", waveform=drag_pair)
    return qp


@pytest.fixture(name="sync_with_dynamic_wait")
def fixture_sync_with_dynamic_wait() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp = QProgram()
    duration = qp.variable(label="time", domain=Domain.Time)
    with qp.for_loop(variable=duration, start=100, stop=200, step=10):
        qp.play(bus="drive", waveform=drag_pair)
        qp.wait(bus="drive", duration=duration)
        qp.sync()
        qp.play(bus="readout", waveform=drag_pair)
    return qp


@pytest.fixture(name="infinite_loop")
def fixture_infinite_loop() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp = QProgram()
    with qp.infinite_loop():
        qp.play(bus="drive", waveform=drag_pair)
    return qp


@pytest.fixture(name="average_loop")
def fixture_average_loop() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights = IQPair(
        I=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
        Q=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
    )
    qp = QProgram()
    with qp.average(shots=1000):
        qp.play(bus="drive", waveform=drag_pair)
        qp.sync()
        qp.wait(bus="readout", duration=100)
        qp.play(bus="readout", waveform=readout_pair)
        qp.qblox.acquire(bus="readout", weights=weights)
    return qp


@pytest.fixture(name="average_loop_long_wait")
def fixture_average_loop_long_wait() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights = IQPair(
        I=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
        Q=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
    )
    qp = QProgram()
    with qp.average(shots=1000):
        qp.play(bus="drive", waveform=drag_pair)
        qp.wait(bus="drive", duration=100_000)
        qp.sync()
        qp.play(bus="readout", waveform=readout_pair)
        qp.qblox.acquire(bus="readout", weights=weights)
    return qp


@pytest.fixture(name="acquire_with_weights_of_different_length")
def fixture_acquire_with_weights_of_different_lengths() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights = IQPair(
        I=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
        Q=Gaussian(amplitude=1.0, duration=500, num_sigmas=2.5),
    )
    qp = QProgram()
    with qp.average(shots=1000):
        qp.play(bus="drive", waveform=drag_pair)
        qp.sync()
        qp.wait(bus="readout", duration=100)
        qp.play(bus="readout", waveform=readout_pair)
        qp.qblox.acquire(bus="readout", weights=weights)
    return qp


@pytest.fixture(name="average_with_for_loop")
def fixture_average_with_for_loop() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.average(shots=1000):
        with qp.for_loop(variable=gain, start=0, stop=1.0, step=0.1):
            qp.play(bus="drive", waveform=drag_pair)
            qp.set_gain(bus="readout", gain=gain)
            qp.play(bus="readout", waveform=readout_pair)
            qp.qblox.acquire(bus="readout", weights=weights_pair)
    return qp


@pytest.fixture(name="average_with_for_loop_nshots")
def fixture_average_with_for_loop_nshots() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    nshots = qp.variable(label="nshots", domain=Domain.Scalar, type=int)
    with qp.average(shots=1000):
        with qp.for_loop(variable=nshots, start=0, stop=2, step=1):
            qp.play(bus="drive", waveform=drag_pair)
            qp.play(bus="readout", waveform=readout_pair)
            qp.qblox.acquire(bus="readout", weights=weights_pair)
    return qp


@pytest.fixture(name="acquire_loop_with_for_loop_with_weights_of_same_waveform")
def fixture_acquire_loop_with_for_loop_with_weights_of_same_waveform() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights = IQPair(
        I=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
        Q=Gaussian(amplitude=1.0, duration=1000, num_sigmas=2.5),
    )
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.average(shots=1000):
        with qp.for_loop(variable=gain, start=0, stop=1.0, step=0.1):
            qp.play(bus="drive", waveform=drag_pair)
            qp.set_gain(bus="readout", gain=gain)
            qp.play(bus="readout", waveform=readout_pair)
            qp.qblox.acquire(bus="readout", weights=weights)
    return qp


@pytest.fixture(name="average_with_multiple_for_loops_and_acquires")
def fixture_average_with_multiple_for_loops_and_acquires() -> QProgram:
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair_0 = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    weights_pair_1 = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair_2 = IQPair(I=Square(amplitude=1.0, duration=500), Q=Square(amplitude=0.0, duration=500))
    qp = QProgram()
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.average(shots=1000):
        with qp.for_loop(variable=frequency, start=0, stop=500, step=10):
            qp.set_frequency(bus="readout", frequency=frequency)
            qp.play(bus="readout", waveform=readout_pair)
            qp.qblox.acquire(bus="readout", weights=weights_pair_0)
        qp.qblox.acquire(bus="readout", weights=weights_pair_1)
        with qp.for_loop(variable=gain, start=0.0, stop=1.0, step=0.1):
            qp.set_gain(bus="readout", gain=gain)
            qp.play(bus="readout", waveform=readout_pair)
            qp.qblox.acquire(bus="readout", weights=weights_pair_2)
    return qp


@pytest.fixture(name="average_with_nested_for_loops")
def fixture_average_with_nested_for_loops() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.average(shots=1000):
        with qp.for_loop(variable=gain, start=0, stop=1, step=0.1):
            qp.set_gain(bus="drive", gain=gain)
            with qp.for_loop(variable=frequency, start=0, stop=500, step=10):
                qp.play(bus="drive", waveform=drag_pair)
                qp.sync()
                qp.set_frequency(bus="readout", frequency=frequency)
                qp.play(bus="readout", waveform=readout_pair)
                qp.qblox.acquire(bus="readout", weights=weights_pair)
    return qp


@pytest.fixture(name="measure_program")
def fixture_measure_program() -> QProgram:
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    qp.measure(bus="readout", waveform=readout_pair, weights=weights_pair)
    return qp


@pytest.fixture(name="average_with_parallel_for_loops")
def fixture_average_with_parallel_for_loops() -> QProgram:
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
    weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
    qp = QProgram()
    frequency = qp.variable(label="frequency", domain=Domain.Frequency)
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.average(shots=1000):
        with qp.parallel(
            loops=[
                ForLoop(variable=frequency, start=100, stop=200, step=10),
                ForLoop(variable=gain, start=0, stop=1, step=0.1),
            ]
        ):
            qp.set_gain(bus="drive", gain=gain)
            qp.set_frequency(bus="readout", frequency=frequency)
            qp.play(bus="drive", waveform=drag_pair)
            qp.sync()
            qp.play(bus="readout", waveform=readout_pair)
            qp.qblox.acquire(bus="readout", weights=weights_pair)
    return qp


@pytest.fixture(name="for_loop_variable_with_no_target")
def fixture_for_loop_variable_with_no_target() -> QProgram:
    qp = QProgram()
    variable = qp.variable(label="float_scalar", domain=Domain.Scalar, type=float)
    with qp.average(shots=1000):
        with qp.for_loop(variable=variable, start=0, stop=100, step=4):
            qp.set_frequency(bus="drive", frequency=100)
            qp.set_phase(bus="drive", phase=90)
    return qp


@pytest.fixture(name="play_operation_with_waveforms_of_different_length")
def fixture_play_operation_with_waveforms_of_different_length() -> QProgram:
    qp = QProgram()
    waveform_pair = IQPair(I=Square(amplitude=1.0, duration=40), Q=Square(amplitude=1.0, duration=80))
    qp.play(bus="drive", waveform=waveform_pair)
    return qp


@pytest.fixture(name="multiple_play_operations_with_same_waveform")
def fixture_multiple_play_operations_with_same_waveform() -> QProgram:
    qp = QProgram()
    drag_pair = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2)
    qp.play(bus="drive", waveform=drag_pair)
    qp.play(bus="drive", waveform=drag_pair)
    qp.play(bus="drive", waveform=IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4, drag_coefficient=1.2))
    return qp


@pytest.fixture(name="multiple_play_operations_with_no_Q_waveform")
def fixture_multiple_play_operations_with_no_Q_waveform() -> QProgram:
    qp = QProgram()
    gaussian = Gaussian(amplitude=1.0, duration=40, num_sigmas=4)
    qp.play(bus="drive", waveform=gaussian)
    qp.play(bus="drive", waveform=gaussian)
    qp.play(bus="drive", waveform=Gaussian(amplitude=1.0, duration=40, num_sigmas=4))
    return qp

@pytest.fixture(name="play_square_waveforms_with_optimization")
def fixture_lay_square_waveforms_with_optimization() -> QProgram:
    qp = QProgram()
    qp.play(bus="drive", waveform=IQPair(I=Square(1.0, duration=25), Q=Square(0.5, duration=25)))
    qp.play(bus="drive", waveform=Square(1.0, duration=50))
    qp.play(bus="drive", waveform=Square(1.0, duration=500))
    qp.play(bus="drive", waveform=IQPair(I=Square(1.0, duration=500), Q=Square(0.0, duration=500)))
    qp.play(bus="drive", waveform=Square(1.0, duration=50_000))
    qp.play(bus="drive", waveform=Square(1.0, duration=9790223))
    qp.play(bus="drive", waveform=IQPair(I=Square(1.0, duration=9790223), Q=Square(1.0, duration=9790223)))
    qp.play(bus="drive", waveform=Square(0.5, duration=1234567))
    return qp

@pytest.fixture(name="play_operation_with_variable_in_waveform")
def fixture_play_operation_with_variable_in_waveform() -> QProgram:
    qp = QProgram()
    amplitude = qp.variable(label="amplitude", domain=Domain.Voltage)
    qp.play(bus="drive", waveform=Square(amplitude=amplitude, duration=100))
    return qp

@pytest.fixture(name="update_latched_param")
def update_latched_param() -> QProgram:
    qp = QProgram()
    qp.set_offset("drive",1,0)
    qp.wait(bus="drive", duration=0)
    qp.play(bus="drive", waveform=Square(amplitude=1, duration=100))
    qp.set_phase("drive",1)
    qp.wait(bus="drive", duration=4)
    qp.play(bus="drive", waveform=Square(amplitude=1, duration=100))
    qp.set_gain("drive",1)
    qp.wait(bus="drive", duration=100)
    qp.play(bus="drive", waveform=Square(amplitude=1, duration=5))
    qp.set_frequency("drive",1e6)
    qp.wait(bus="drive", duration=100000)
    qp.play(bus="drive", waveform=Square(amplitude=1, duration=5))
    qp.set_gain("drive",1)
    qp.wait(bus="drive", duration=4)
    qp.play(bus="drive", waveform=Square(amplitude=1, duration=5))
    qp.set_offset("drive",1,0)
    qp.wait(bus="drive", duration=6)
    return qp

class TestQBloxCompiler:
    def test_play_named_operation_and_bus_mapping(self, play_named_operation: QProgram, calibration: Calibration):
        compiler = QbloxCompiler()
        output = compiler.compile(
            qprogram=play_named_operation, bus_mapping={"drive": "drive_q0"}, calibration=calibration
        )

        assert len(output.sequences) == 1
        assert "drive_q0" in output.sequences
        assert isinstance(output.sequences["drive_q0"], QPy.Sequence)

    def test_block_handlers(self, measurement_blocked_operation: QProgram, calibration: Calibration):
        drag_wf = IQPair.DRAG(amplitude=1.0, duration=100, num_sigmas=5, drag_coefficient=1.5)
        readout_pair = IQPair(I=Square(amplitude=1.0, duration=1000), Q=Square(amplitude=0.0, duration=1000))
        weights_pair = IQPair(I=Square(amplitude=1.0, duration=2000), Q=Square(amplitude=0.0, duration=2000))
        qp_no_block = QProgram()
        qp_no_block.play(bus="drive", waveform=drag_wf)
        qp_no_block.measure(bus="readout", waveform=readout_pair, weights=weights_pair)

        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(
            qprogram=measurement_blocked_operation
        )

        sequences_no_block, _ = compiler.compile(
            qprogram=qp_no_block
        )
        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        drive_str = """
                setup:
                                wait_sync        4
                                set_mrk          0
                                upd_param        4

                main:
                                play             0, 1, 100
                                set_mrk          0
                                upd_param        4
                                stop
            """
        assert is_q1asm_equal(sequences["drive"]._program, drive_str)
        assert is_q1asm_equal(sequences["drive"]._program, sequences_no_block["drive"]._program)

    def test_play_named_operation_raises_error_if_operations_not_in_calibration(self, play_named_operation: QProgram):
        calibration = Calibration()
        compiler = QbloxCompiler()
        with pytest.raises(RuntimeError):
            _ = compiler.compile(play_named_operation, bus_mapping={"drive": "drive_q0"}, calibration=calibration)

    def test_no_loops_all_operations(self, no_loops_all_operations: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=no_loops_all_operations)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            set_freq         1200
                            set_freq         1200
                            set_ph           250000000
                            reset_ph
                            set_awg_gain     16383, 16383
                            set_awg_gain     16383, 16383
                            set_awg_offs     16383, 16383
                            play             0, 1, 40
                            set_mrk          0
                            upd_param        4
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)

        assert len(sequences["readout"]._waveforms._waveforms) == 4
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 1
        assert len(sequences["readout"]._weights._weights) == 2
        assert sequences["readout"]._program._compiled

        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            wait             40         
                            wait             100            
                            move             10, R0         
            square_0:
                            play             0, 1, 100      
                            loop             R0, @square_0  
                            set_mrk          7              
                            play             2, 3, 4        
                            acquire_weighed  0, 0, 0, 1, 2000
                            set_mrk          0              
                            upd_param        4              
                            stop  
        """
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_set_offset_without_path_1_throws_exception(self, caplog, offset_no_path1: QProgram):
        compiler = QbloxCompiler()
        with caplog.at_level(logging.WARNING):
            _ = compiler.compile(qprogram=offset_no_path1)
        assert "Qblox requires an offset for the two paths, the offset of the second path has been set to the same as the first path." in caplog.text

    def test_dynamic_wait(self, dynamic_wait: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=dynamic_wait)

        assert len(sequences) == 1
        assert "drive" in sequences

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             11, R0
                            move             100, R1
            loop_0:
                            play             0, 1, 40
                            wait             R1
                            add              R1, 10, R1
                            loop             R0, @loop_0
                            set_mrk          0
                            upd_param        4
                            stop
        """

        assert is_q1asm_equal(sequences["drive"], drive_str)

    def test_dynamic_wait_multiple_buses_with_disable_autosync(
        self, dynamic_wait_multiple_buses_with_disable_autosync: QProgram
    ):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=dynamic_wait_multiple_buses_with_disable_autosync)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             11, R0
                            move             100, R1
            loop_0:
                            play             0, 1, 40
                            add              R1, 10, R1
                            loop             R0, @loop_0
                            set_mrk          0
                            upd_param        4
                            stop
        """

        readout_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             11, R0
                            move             100, R1
            loop_0:
                            wait             R1
                            play             0, 1, 40
                            add              R1, 10, R1
                            loop             R0, @loop_0
                            nop
                            set_mrk          0
                            upd_param        4
                            stop
        """

        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_dynamic_wait_multiple_buses_throws_exception(self, dynamic_wait_multiple_buses: QProgram):
        with pytest.raises(NotImplementedError, match="Dynamic syncing is not implemented yet."):
            compiler = QbloxCompiler()
            _ = compiler.compile(qprogram=dynamic_wait_multiple_buses)

    def test_sync_operation_with_dynamic_timings_throws_exception(self, sync_with_dynamic_wait: QProgram):
        with pytest.raises(NotImplementedError, match="Dynamic syncing is not implemented yet."):
            compiler = QbloxCompiler()
            _ = compiler.compile(qprogram=sync_with_dynamic_wait)

    def test_average_with_long_wait(self, average_loop_long_wait: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_loop_long_wait)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 1
        assert len(sequences["readout"]._weights._weights) == 1
        assert sequences["readout"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4
            main:
                            move             1000, R0
            avg_0:
                            play             0, 1, 40
                            wait             65532
                            wait             34468
                            wait             2000
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """

        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            wait             65532     
                            wait             34508          
                            move             10, R1         
            square_0:
                            play             0, 1, 100      
                            loop             R1, @square_0  
                            acquire_weighed  0, 0, 0, 0, 1000
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop 
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_infinite_loop(self, infinite_loop: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=infinite_loop)

        assert len(sequences) == 1
        assert "drive" in sequences

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
            infinite_loop_0:
                            play             0, 1, 40
                            jmp              @infinite_loop_0
                            set_mrk          0
                            upd_param        4
                            stop
        """

        assert is_q1asm_equal(sequences["drive"], drive_str)

    def test_average_loop(self, average_loop: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_loop)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            play             0, 1, 40
                            wait             2100
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 1
        assert len(sequences["readout"]._weights._weights) == 1
        assert sequences["readout"]._program._compiled

        readout_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            wait             40
                            wait             100
                            move             10, R1
            square_0:
                            play             0, 1, 100      
                            loop             R1, @square_0  
                            acquire_weighed  0, 0, 0, 0, 1000
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_average_with_for_loop_variable_does_nothing(self, average_with_for_loop_nshots: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_for_loop_nshots)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 3
        assert len(sequences["readout"]._weights._weights) == 2
        assert sequences["readout"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             3, R1
                            move             0, R2
            loop_0:
                            play             0, 1, 40
                            wait             2960
                            add              R2, 1, R2
                            loop             R1, @loop_0
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                wait_sync        4              
                set_mrk          0              
                upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             1, R1          
                            move             0, R2          
                            move             0, R3          
                            move             3, R4          
                            move             0, R5          
            loop_0:
                            move             10, R6         
            square_0:
                            play             0, 1, 100      
                            loop             R6, @square_0  
                            acquire_weighed  0, R3, R2, R1, 2000
                            add              R3, 1, R3      
                            add              R5, 1, R5      
                            loop             R4, @loop_0    
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_average_with_for_loop(self, average_with_for_loop: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_for_loop)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 11
        assert len(sequences["readout"]._weights._weights) == 2
        assert sequences["readout"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             11, R1
                            move             0, R2
            loop_0:
                            play             0, 1, 40
                            wait             2960
                            add              R2, 3276, R2
                            loop             R1, @loop_0
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             1, R1          
                            move             0, R2          
                            move             0, R3          
                            move             11, R4         
                            move             0, R5          
            loop_0:
                            set_awg_gain     R5, R5
                            set_awg_gain     R5, R5         
                            move             10, R6         
            square_0:
                            play             0, 1, 100      
                            loop             R6, @square_0  
                            acquire_weighed  0, R3, R2, R1, 2000
                            add              R3, 1, R3      
                            add              R5, 3276, R5   
                            loop             R4, @loop_0    
                            nop                             
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_measure_calls_play_acquire(self, measure_program):
        compiler = QbloxCompiler()

        # Test measure with default time of flight
        with (
            patch.object(QbloxCompiler, "_handle_play") as handle_play,
            patch.object(QbloxCompiler, "_handle_acquire") as handle_acquire,
        ):
            compiler.compile(measure_program)

            measure = measure_program.body.elements[0]
            assert handle_play.call_count == 1
            assert handle_acquire.call_count == 1
            assert handle_play.call_args[0][0].bus == measure.bus
            assert handle_play.call_args[0][0].waveform == measure.waveform
            assert handle_play.call_args[0][0].wait_time == QbloxCompiler.minimum_wait_duration
            assert handle_acquire.call_args[0][0].bus == measure.bus
            assert handle_acquire.call_args[0][0].weights == measure.weights

        # Test measure with provided time of flight
        with (
            patch.object(QbloxCompiler, "_handle_play") as handle_play,
            patch.object(QbloxCompiler, "_handle_acquire") as handle_acquire,
        ):
            compiler.compile(measure_program, times_of_flight={"readout": 123})

            measure = measure_program.body.elements[0]
            assert handle_play.call_count == 1
            assert handle_acquire.call_count == 1
            assert handle_play.call_args[0][0].bus == measure.bus
            assert handle_play.call_args[0][0].waveform == measure.waveform
            assert handle_play.call_args[0][0].wait_time == 123
            assert handle_acquire.call_args[0][0].bus == measure.bus
            assert handle_acquire.call_args[0][0].weights == measure.weights

    def test_acquire_loop_with_for_loop_with_weights_of_same_waveform(
        self, acquire_loop_with_for_loop_with_weights_of_same_waveform: QProgram
    ):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=acquire_loop_with_for_loop_with_weights_of_same_waveform)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 11
        assert len(sequences["readout"]._weights._weights) == 1
        assert sequences["readout"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             11, R1
                            move             0, R2
            loop_0:
                            play             0, 1, 40
                            wait             1960
                            add              R2, 3276, R2
                            loop             R1, @loop_0
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             0, R1          
                            move             0, R2          
                            move             0, R3          
                            move             11, R4         
                            move             0, R5          
            loop_0:
                            set_awg_gain     R5, R5
                            set_awg_gain     R5, R5         
                            move             10, R6         
            square_0:
                            play             0, 1, 100      
                            loop             R6, @square_0  
                            acquire_weighed  0, R3, R2, R1, 1000
                            add              R3, 1, R3      
                            add              R5, 3276, R5   
                            loop             R4, @loop_0    
                            nop                             
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop    
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_average_with_multiple_for_loops_and_acquires(self, average_with_multiple_for_loops_and_acquires: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_multiple_for_loops_and_acquires)

        assert len(sequences) == 1
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 3
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 51
        assert sequences["readout"]._acquisitions._acquisitions[1].num_bins == 1
        assert sequences["readout"]._acquisitions._acquisitions[2].num_bins == 11
        assert len(sequences["readout"]._weights._weights) == 6
        assert sequences["readout"]._program._compiled

        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             5, R1          
                            move             4, R2          
                            move             0, R3          
                            move             1, R4          
                            move             0, R5          
                            move             0, R6          
                            move             51, R7         
                            move             0, R8          
            loop_0:
                            set_freq         R8             
                            set_freq         R8
                            move             10, R9         
            square_0:
                            play             0, 1, 100      
                            loop             R9, @square_0  
                            acquire_weighed  0, R6, R5, R4, 2000
                            add              R6, 1, R6      
                            add              R8, 40, R8     
                            loop             R7, @loop_0    
                            nop                             
                            acquire_weighed  1, 0, 2, 3, 1000
                            move             11, R10        
                            move             0, R11         
                            nop                             
            loop_1:
                            set_awg_gain     R11, R11
                            set_awg_gain     R11, R11       
                            move             10, R12         
            square_1:
                            play             0, 1, 100      
                            loop             R12, @square_1 
                            acquire_weighed  2, R3, R2, R1, 500
                            add              R3, 1, R3      
                            add              R11, 3276, R11 
                            loop             R10, @loop_1   
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop
        """
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_average_with_nested_for_loops(self, average_with_nested_for_loops: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_nested_for_loops)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 561
        assert len(sequences["readout"]._weights._weights) == 2
        assert sequences["readout"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             11, R1
                            move             0, R2
            loop_0:
                            set_awg_gain     R2, R2
                            set_awg_gain     R2, R2
                            move             51, R3
                            move             0, R4
            loop_1:
                            play             0, 1, 40
                            wait             3000
                            add              R4, 40, R4
                            loop             R3, @loop_1
                            add              R2, 3276, R2
                            loop             R1, @loop_0
                            nop
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             1, R1          
                            move             0, R2          
                            move             0, R3          
                            move             11, R4         
                            move             0, R5          
            loop_0:
                            move             51, R6         
                            move             0, R7          
            loop_1:
                            wait             40             
                            set_freq         R7
                            set_freq         R7             
                            move             10, R8         
            square_0:
                            play             0, 1, 100      
                            loop             R8, @square_0  
                            acquire_weighed  0, R3, R2, R1, 2000
                            add              R3, 1, R3      
                            add              R7, 40, R7     
                            loop             R6, @loop_1    
                            add              R5, 3276, R5   
                            loop             R4, @loop_0    
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop   
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_average_with_parallel_for_loops(self, average_with_parallel_for_loops: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_parallel_for_loops)

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        assert len(sequences["readout"]._waveforms._waveforms) == 2
        assert len(sequences["readout"]._acquisitions._acquisitions) == 1
        assert sequences["readout"]._acquisitions._acquisitions[0].num_bins == 11
        assert len(sequences["readout"]._weights._weights) == 2
        assert sequences["readout"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             11, R1
                            move             400, R2
                            move             0, R3
            loop_0:
                            set_awg_gain     R3, R3
                            set_awg_gain     R3, R3
                            play             0, 1, 40
                            wait             3000
                            add              R2, 40, R2
                            add              R3, 3276, R3
                            loop             R1, @loop_0
                            nop
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             1, R1          
                            move             0, R2          
                            move             0, R3          
                            move             11, R4         
                            move             400, R5        
                            move             0, R6          
            loop_0:
                            set_freq         R5
set_freq         R5                            
                            upd_param        4           
                            wait             36             
                            move             10, R7         
            square_0:
                            play             0, 1, 100      
                            loop             R7, @square_0  
                            acquire_weighed  0, R3, R2, R1, 2000
                            add              R3, 1, R3      
                            add              R5, 40, R5     
                            add              R6, 3276, R6   
                            loop             R4, @loop_0    
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop 
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_multiple_play_operations_with_same_waveform(self, multiple_play_operations_with_same_waveform: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=multiple_play_operations_with_same_waveform)

        assert len(sequences) == 1
        assert "drive" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            play             0, 1, 40
                            play             0, 1, 40
                            play             0, 1, 40
                            set_mrk          0
                            upd_param        4
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)

    def test_multiple_play_operations_with_no_Q_waveform(self, multiple_play_operations_with_no_Q_waveform: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=multiple_play_operations_with_no_Q_waveform)

        assert len(sequences) == 1
        assert "drive" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 2
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            play             0, 1, 40
                            play             0, 1, 40
                            play             0, 1, 40
                            set_mrk          0
                            upd_param        4
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)

    def test_play_square_waveforms_with_optimization(self, play_square_waveforms_with_optimization: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=play_square_waveforms_with_optimization)

        assert len(sequences["drive"]._waveforms._waveforms) == 11
        assert sequences["drive"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            play             0, 1, 25
                            play             2, 3, 50
                            move             5, R0
            square_0:
                            play             4, 5, 100
                            loop             R0, @square_0
                            move             5, R1
            square_1:
                            play             4, 6, 100
                            loop             R1, @square_1
                            move             500, R2
            square_2:
                            play             4, 5, 100
                            loop             R2, @square_2
                            move             97902, R3
            square_3:
                            play             4, 5, 100
                            loop             R3, @square_3
                            play             7, 8, 23
                            move             97902, R4
            square_4:
                            play             4, 4, 100
                            loop             R4, @square_4
                            play             7, 7, 23
                            move             9721, R5
            square_5:
                            play             9, 10, 127
                            loop             R5, @square_5
                            set_mrk          0
                            upd_param        4
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)

    def test_play_operation_with_variable_in_waveform(self, caplog, play_operation_with_variable_in_waveform: QProgram):
        compiler = QbloxCompiler()
        with caplog.at_level(logging.ERROR):
            _ = compiler.compile(qprogram=play_operation_with_variable_in_waveform)

        assert "Variables in waveforms are not supported in Qblox." in caplog.text


    def test_delay(self, average_with_for_loop_nshots: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_for_loop_nshots, delays={"drive": 20})

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             3, R1
                            move             0, R2
            loop_0:
                            wait             20
                            play             0, 1, 40
                            wait             2960
                            add              R2, 1, R2
                            loop             R1, @loop_0
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             1, R1          
                            move             0, R2          
                            move             0, R3          
                            move             3, R4          
                            move             0, R5          
            loop_0:
                            move             10, R6         
            square_0:
                            play             0, 1, 100      
                            loop             R6, @square_0  
                            acquire_weighed  0, R3, R2, R1, 2000
                            add              R3, 1, R3     
                            wait             20             
                            add              R5, 1, R5      
                            loop             R4, @loop_0    
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    def test_negative_delay(self, average_with_for_loop_nshots: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=average_with_for_loop_nshots, delays={"drive": -20})

        assert len(sequences) == 2
        assert "drive" in sequences
        assert "readout" in sequences

        drive_str = """
            setup:
                            wait_sync        4
                            set_mrk          0
                            upd_param        4

            main:
                            move             1000, R0
            avg_0:
                            move             3, R1
                            move             0, R2
            loop_0:
                            play             0, 1, 40
                            wait             2980
                            add              R2, 1, R2
                            loop             R1, @loop_0
                            loop             R0, @avg_0
                            set_mrk          0
                            upd_param        4
                            stop
        """
        readout_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            move             1000, R0       
            avg_0:
                            move             1, R1          
                            move             0, R2          
                            move             0, R3          
                            move             3, R4          
                            move             0, R5          
            loop_0:
                            wait             20             
                            move             10, R6         
            square_0:
                            play             0, 1, 100      
                            loop             R6, @square_0  
                            acquire_weighed  0, R3, R2, R1, 2000
                            add              R3, 1, R3      
                            add              R5, 1, R5      
                            loop             R4, @loop_0    
                            loop             R0, @avg_0     
                            set_mrk          0              
                            upd_param        4              
                            stop
        """
        assert is_q1asm_equal(sequences["drive"], drive_str)
        assert is_q1asm_equal(sequences["readout"], readout_str)

    @pytest.mark.parametrize(
        "start,stop,step,expected_result",
        [(0, 10, 1, 11), (10, 0, -1, 11), (1, 2.05, 0.1, 11)],
    )
    def test_calculate_iterations(self, start, stop, step, expected_result):
        result = QbloxCompiler._calculate_iterations(start, stop, step)
        assert result == expected_result

    def test_calculate_iterations_with_zero_step_throws_error(self):
        with pytest.raises(ValueError, match="Step value cannot be zero"):
            QbloxCompiler._calculate_iterations(100, 200, 0)

    def test_update_latched_param_before_wait(self, update_latched_param: QProgram):
        compiler = QbloxCompiler()
        sequences, _ = compiler.compile(qprogram=update_latched_param)

        assert len(sequences) == 1
        assert "drive" in sequences

        for bus in sequences:
            assert isinstance(sequences[bus], QPy.Sequence)

        assert len(sequences["drive"]._waveforms._waveforms) == 4
        assert len(sequences["drive"]._acquisitions._acquisitions) == 0
        assert len(sequences["drive"]._weights._weights) == 0
        assert sequences["drive"]._program._compiled

        drive_str = """
            setup:
                            wait_sync        4              
                            set_mrk          0              
                            upd_param        4              

            main:
                            set_awg_offs     32767, 0       
                            upd_param        4              
                            move             1, R0          
            square_0:
                            play             0, 1, 100      
                            loop             R0, @square_0  
                            set_ph           159154943      
                            upd_param        4              
                            move             1, R1          
            square_1:
                            play             0, 1, 100      
                            loop             R1, @square_1  
                            set_awg_gain     32767, 32767
                            set_awg_gain     32767, 32767   
                            upd_param        4              
                            wait             96             
                            play             2, 3, 5        
                            set_freq         4000000
                            set_freq         4000000        
                            upd_param        4              
                            wait             65532          
                            wait             34464          
                            play             2, 3, 5        
                            set_awg_gain     32767, 32767
                            set_awg_gain     32767, 32767   
                            upd_param        4              
                            play             2, 3, 5        
                            set_awg_offs     32767, 0       
                            upd_param        6              
                            set_mrk          0              
                            upd_param        4              
                            stop                            
        """
        
        assert is_q1asm_equal(sequences["drive"], drive_str)


"""Test StreamArray"""

import copy
from unittest.mock import patch

import numpy as np
import pytest

from qililab.result import stream_results
from qililab.result.stream_results import StreamArray

AMP_VALUES = np.arange(0, 1, 2)


@pytest.fixture(name="stream_array")
def fixture_stream_array():
    """fixture_stream_array

    Returns:
        stream_array: StreamArray
    """
    shape = (2, 2)
    path = "test_stream_array.hdf5"
    loops = {"test_amp_loop": AMP_VALUES}

    return stream_results(shape=shape, path=path, loops=loops)


class MockGroup:
    """Mock a h5py group."""

    def create_dataset(self, name: str, data: np.ndarray):
        """Creates a dataset"""
        return {}


class MockFile:
    """Mocks a h5py file."""

    def __init__(self):
        """Initialize a mock file."""
        self.dataset = None

    def create_group(self, name: str):
        return MockGroup()

    def create_dataset(self, _: str, data: np.ndarray):
        """Creates a dataset"""
        return copy.deepcopy(data)

    def __exit__(self, *_):
        """mocks exit"""


class TestStreamArray:
    """Test `StreamArray` functionalities."""

    def test_stream_array_instantiation(self, stream_array: StreamArray):
        """Tests the instantiation of a StreamArray object."""
        assert stream_array.results.shape == (2, 2)
        assert (stream_array.results == np.empty(shape=(2, 2))).all
        assert stream_array.path == "test_stream_array.hdf5"
        assert stream_array.loops == {"test_amp_loop": np.arange(0, 1, 2)}

    @patch("h5py.File", return_value=MockFile())
    def test_context_manager(self, mock_h5py: MockFile, stream_array: StreamArray):
        """Tests context manager real time saving."""
        # test adding outside the context manager
        stream_array[0, 0] = -2

        assert stream_array._dataset is None

        # test adding inside the context manager
        with stream_array:
            stream_array[0, 0] = 1
            stream_array[0, 1] = 2
            stream_array[1, 0] = 3
            stream_array[1, 1] = 4

        assert (stream_array.results == [[1, 2], [3, 4]]).all
        assert stream_array._dataset is not None
        assert (stream_array._dataset == [[1, 2], [3, 4]]).all
        assert stream_array._dataset[0, 0] == 1

        assert len(stream_array) == 2
        assert sum(1 for _ in iter(stream_array)) == 2
        assert str(stream_array) == "[[1. 2.]\n [3. 4.]]"

        assert [1, 2] in stream_array
        assert (stream_array[0] == [1, 2]).all



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Drag pulse shape."""

from dataclasses import dataclass

import numpy as np

from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import PulseShapeName
from qililab.utils import Factory


@Factory.register
@dataclass(frozen=True, eq=True)
class Drag(PulseShape):
    """Derivative Removal by Adiabatic Gate (DRAG) pulse shape. Standard Gaussian pulse with an additional Gaussian derivative component.

    It is designed to reduce the frequency spectrum of a normal gaussian pulse near the :math:`|1> - |2>` transition, reducing the chance of leakage to the :math:`|2>` state:

    .. math::

        f(x) & = (1 + 1j * alpha * d/dx) * Gaussian(x) = \\\\
             & = (1 + 1j * alpha * (x - \\mu) / \\sigma^2) * Gaussian(x)

    where the 'Gaussian' is symmetrically cut in the given `num_sigmas`, and then is shifted down so that the extremes are at 0, following:

    .. math::

        Gaussian(x) = A * exp(-0.5 * (x - \\mu)^2 / \\sigma^2)

    more information about the gaussian can be found in the :class:`Gaussian` documentation.

    Examples:
        To get the envelope of a drag shape, with ``num_sigmas = 8`` and ``drag_coefficient`` (`alpha`) equal to ``X``, you need to do:

        .. code-block:: python

            from qililab.pulse.pulse_shape import Drag
            rectangular_envelope = Drag(num_sigmas=8, drag_coefficient=X).envelope(amplitude=1., duration=50)

        which for ``X`` being ``1`` and ``10``, look respectively like:

        .. image:: /classes_images/drags.png
            :width: 800
            :align: center

    References:
        - Analytic control methods for high-fidelity unitary operations in a weakly nonlinear oscillator: https://journals.aps.org/pra/abstract/10.1103/PhysRevA.83.012308.
        - Simple Pulses for Elimination of Leakage in Weakly Nonlinear Qubits: https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.103.110501.

    Args:
        num_sigmas (float): Sigma number of the gaussian pulse shape. Defines the width of the gaussian pulse.
        drag_coefficient (float): Drag coefficient that gives the DRAG its imaginary components.
    """

    name = PulseShapeName.DRAG  # noqa: RUF045
    num_sigmas: float
    drag_coefficient: float

    def envelope(self, duration: int, amplitude: float, resolution: float = 1.0) -> np.ndarray:
        """DRAG envelope centered with respect to the pulse.

        The first point of the gaussian in the envelope is shifted to avoid introducing noise at time 0.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """
        sigma = duration / self.num_sigmas
        time = np.arange(duration / resolution) * resolution
        mu_ = duration / 2

        gaussian = amplitude * np.exp(-0.5 * (time - mu_) ** 2 / sigma**2)
        norm = np.amax(np.abs(np.real(gaussian)))

        gaussian = gaussian - gaussian[0]  # Shift to avoid introducing noise at time 0
        corr_norm = np.amax(np.abs(np.real(gaussian)))

        drag_gaussian = (1 - 1j * self.drag_coefficient * (time - mu_) / sigma**2) * gaussian

        return drag_gaussian * norm / corr_norm if corr_norm != 0 else drag_gaussian



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import asdict, dataclass, field

from qililab.pulse.pulse_distortion import PulseDistortion
from qililab.typings.enums import Line
from qililab.utils.castings import cast_enum_fields


@dataclass
class DigitalCompilationBusSettings:
    """Settings for a single gate event. A gate event is an element of a gate schedule, which is the
    sequence of gate events that define what a gate does (the pulse events it consists of).

    A gate event is made up of GatePulseSettings, which contains pulse specific information, and extra arguments
    (bus and wait_time) which are related to the context in which the specific pulse in GatePulseSettings is applied

    Attributes:
        bus (str): bus through which the pulse is to be sent. The string has to match that of the bus alias in the runcard
        pulse (GatePulseSettings): settings of the bus to be launched
        wait_time (int): time to wait w.r.t gate start time (taken as 0) before launching the pulse
    """

    line: Line
    qubits: list[int]
    delay: int = 0
    distortions: list[PulseDistortion] = field(default_factory=list)

    def __post_init__(self):
        cast_enum_fields(obj=self)
        self.distortions = [
            PulseDistortion.from_dict(distortion)
            for distortion in self.distortions
            if isinstance(distortion, dict)  # type: ignore[arg-type]
        ]

    def is_readout(self):
        """Return true if bus is readout."""
        return self.line == Line.READOUT

    def to_dict(self):
        return asdict(self) | {"distortions": [distortion.to_dict() for distortion in self.distortions]}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Cluster Controller class"""

from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentController, InstrumentControllerSettings
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.qblox.qblox_qcm import QbloxQCM
from qililab.instruments.qblox.qblox_qrm import QbloxQRM
from qililab.typings.enums import (
    ConnectionName,
    InstrumentControllerName,
    InstrumentTypeName,
    Parameter,
    ReferenceClock,
)
from qililab.typings.instruments.cluster import Cluster


@InstrumentControllerFactory.register
class QbloxClusterController(InstrumentController):
    """Qblox Cluster Controller class.

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        number_available_modules (int): Number of modules available in the Instrument Controller.
        settings (QbloxClusterControllerSettings): Settings of the Qblox Pulser Instrument Controller.
    """

    name = InstrumentControllerName.QBLOX_CLUSTER
    number_available_modules = 20
    device: Cluster
    modules: Sequence[QbloxQCM | QbloxQRM]

    @dataclass
    class QbloxClusterControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific Qblox Cluster Controller."""

        reference_clock: ReferenceClock

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.TCP_IP

    settings: QbloxClusterControllerSettings

    @InstrumentController.CheckConnected
    def initial_setup(self):
        """Initial setup of the Qblox Cluster Controller."""
        self._set_reference_source()
        super().initial_setup()

    @InstrumentController.CheckConnected
    def reset(self):
        """Reset the device and clear cache of all modules."""
        self.device.reset()
        for module in self.modules:
            module.clear_cache()

    @InstrumentController.CheckConnected
    def _set_reference_source(self):
        """Set the reference source ('internal' or 'external')."""
        self.device.reference_source(self.reference_clock.value)

    @property
    def reference_clock(self):
        """Get the reference clock setting."""
        return self.settings.reference_clock

    def _check_supported_modules(self):
        """Check if all loaded instrument modules are supported."""
        for module in self.modules:
            if not isinstance(module, (QbloxQCM, QbloxQRM)):
                raise ValueError(
                    f"Instrument {type(module)} not supported. "
                    f"The only supported instruments are {InstrumentTypeName.QBLOX_QCM} and {InstrumentTypeName.QBLOX_QRM}."
                )

    def _initialize_device(self):
        """Initialize the cluster device."""
        self.device = Cluster(name=f"{self.name.value}_{self.alias}", identifier=self.address)

    def _set_device_to_all_modules(self):
        """Set the initialized device to all attached modules."""
        for module, slot_id in zip(self.modules, self.connected_modules_slot_ids):
            module.device = self.device.modules[slot_id - 1]  # slot_id represents the number displayed in the cluster

    def to_dict(self):
        """Return a dictionary representation of the Qblox controller class."""
        return super().to_dict() | {
            Parameter.REFERENCE_CLOCK.value: self.reference_clock.value,
        }



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import Counter
from typing import Counter as TCounter

import numpy as np


class Counts:
    """
    Class representing the number of times (counts) each state was measured.
    """

    def __init__(self, n_qubits: int):
        self.n_qubits = n_qubits
        n_states = 2**self.n_qubits if self.n_qubits > 0 else 0
        # Initialize counter with all n-qubits basis elements as keys and 0 as values.
        self._counter = Counter({bin(i)[2:].zfill(self.n_qubits): 0 for i in range(n_states)})
        self._total_measurements = 0

    @property
    def counter(self) -> TCounter[str]:
        """Counter object holding the counts.

        Returns:
            collections.Counter[str]: Counter object holding the counts.
        """
        return self._counter

    def as_dict(self) -> dict[str, int]:
        """Dictionary representation of the Counts object

        Returns:
            dict[str, int]: Dictionary containing the number of measurements (value) at each state (key).
        """
        return dict(self._counter)

    @property
    def total_measurements(self) -> int:
        """Number of total measurements held by the Counts object.

        Returns:
            int: number of measurements at all the states.
        """
        return self._total_measurements

    def add_measurement(self, state: str):
        """Adds a measurement to the Counts object.

        Args:
            state (str): string with 0s and 1s representing the state to add to the Count.
        """
        self._verify_state(state)
        self._counter.update([state])
        self._total_measurements += 1

    def add_counts(self, counts: Counts):
        """Adds all the counts from another Counts object to this Counts object.

        Args:
            counts (Counts): Counts object to add.

        Raises:
            IndexError: Received a counts object with a different number of qubits.
        """
        if self.n_qubits != counts.n_qubits:
            raise IndexError(
                f"Counts object with {counts.n_qubits} can't be added to Counts object with {self.n_qubits} qubits."
            )
        self._counter.update(counts.counter)
        self._total_measurements += counts.total_measurements

    def probabilities(self) -> dict[str, float]:
        """Returns the probabilities of each measurement.

        Returns:
            dict[str, float]: Dictionary of probabilities (value) of each measurement (key).
        """
        return {measurement: counts / self._total_measurements for (measurement, counts) in self._counter.items()}

    def _verify_state(self, state: str) -> None:
        """Verifies that a state is valid for the Counts object.

        Args:
            state (str): string with 0s and 1s representing the state to add to the Count.

        Raises:
            IndexError: Received a state representing a number of qubits different from the number of qubits of the Counts object.
            ValueError: Received a state made of elements other than 0s and 1s.
        """
        if len(state) != self.n_qubits:
            raise IndexError(f"State '{state}' can't be added to Counts object with {self.n_qubits} qubits.")
        if any(bit not in ["0", "1"] for bit in state):
            raise ValueError(f"A state can only be made of 0s and 1s, got state '{state}'.")

    def __str__(self) -> str:
        """String representation of the Counts object.

        Returns:
            str: String representation of the Counts object.
        """
        return str(self.as_dict())

    def __iadd__(self, other: Counts) -> Counts:
        """In-place addition of two Counts objects (self += other). Adds the `other` Counts object to itself.

        Args:
            other (Counts): Counts object to add to this Counts object.

        Returns:
            Counts: Reference to self.
        """
        self.add_counts(other)
        return self

    @staticmethod
    def from_qprogram(qprogram_results, qubit_mapping: list[str] | None = None):
        """Returns `Count` object from a `QProgramResults`.

        Args:
            qprogram_results (QProgramResults): The QProgramResults object we want to get the `Counts` from.
            qubit_mapping (list[str], optional): A list containing the names of the buses to map.
                Each bus is mapped to a qubit at the corresponding index; the bus at the i-th position in the list is mapped to the i-th qubit.
                Defaults to None.

        Raises:
           ValueError: If the `results` attribute of `qprogram_results` is an empty dictionary.
            ValueError: If the qubit mapping is incomplete and does not map all qubits.
            ValueError: If a qubit mapping is specified and any of the buses do not match the ones in the runcard.

        Returns:
            dict[str, float]: A dictionary where the keys are the quantum states, and the values are the probabilities obtained for each state.
        """
        if not qprogram_results.results:
            raise ValueError(f"Can not obtain counts with no measurments, {qprogram_results.__class__} empty")

        n_qubits = len(qprogram_results.results)
        counts_object = Counts(n_qubits)
        buses = list(qprogram_results.results.keys())

        if qubit_mapping is not None:
            unique_mapping = set(qubit_mapping)  # Remove possible repeated elements
            if n_qubits != len(unique_mapping):
                raise ValueError(
                    f"Expected mapping for all qubits. Results have {n_qubits} qubits, but only {len(unique_mapping)} diferent buses were specified on the mapping."
                )
            if not unique_mapping.issubset(set(buses)):
                raise ValueError(
                    "No measurements found for all specified buses, check the name of the buses provided with the mapping match all the buses specified in runcard."
                )
            buses = qubit_mapping

        # The threshold inside of a qblox bin is the name they use for already classified data as a value between
        # 0 and 1, not the value used in the comparator to perform such classification.
        threshold_matrix = np.array(
            [
                np.concatenate([measurement.threshold.astype(int) for measurement in qprogram_results.results[bus]])
                for bus in buses
            ]
        )
        threshold_matrix_T = threshold_matrix.transpose()
        for state in threshold_matrix_T:
            binary_state_str = "".join(state.astype(str))
            counts_object.add_measurement(state=binary_state_str)

        return counts_object



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MeasurementResult class."""

from qililab.result.qprogram.measurement_result import MeasurementResult
from qililab.yaml import yaml


@yaml.register_class
class QProgramResults:
    """Results from a single execution of QProgram."""

    def __init__(self) -> None:
        self.results: dict[str, list[MeasurementResult]] = {}
        self._timeline: list[MeasurementResult] = []

    @property
    def timeline(self):
        """Retrieve all measurement results in the order they were inserted.

        Returns:
            list[MeasurementResult]: List of MeasurementResults in insertion order.
        """
        return self._timeline

    def append_result(self, bus: str, result: MeasurementResult):
        """Append a measurement result to bus's results list.

        Args:
            bus (str): The bus alias
            result (MeasurementResult): The measurement result to append.
        """
        if bus not in self.results:
            self.results[bus] = []
        self.results[bus].append(result)

        # Append to global insertion order
        self._timeline.append(result)



"""Unit tests for all the methods for data management."""

import copy
import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

import qililab as ql
from qililab.data_management import load_results, save_platform, save_results
from qililab.platform import Platform
from tests.data import Galadriel
from tests.test_utils import build_platform


@patch("ruamel.yaml.YAML.load", return_value=copy.deepcopy(Galadriel.runcard))
@patch("qililab.data_management.open")
class TestPlatformData:
    """Unit tests for the `build_platform` function.."""

    def test_build_platform_passing_a_path_to_runcard_argument(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test build method, with a string path."""
        platform = ql.build_platform(runcard="_")
        assert isinstance(platform, Platform)
        mock_load.assert_called_once()
        mock_open.assert_called_once()

    def test_build_platform_passing_a_dict_to_runcard_argument(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test build method, with a direct dict."""
        platform = ql.build_platform(runcard=copy.deepcopy(Galadriel.runcard))
        assert isinstance(platform, Platform)
        mock_load.assert_not_called()
        mock_open.assert_not_called()

    def test_save_platform_with_non_yml_path(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test the `save_platform` function with a path that doesn't end in .yml."""
        path = save_platform(path="test/", platform=build_platform(Galadriel.runcard))

        mock_open.assert_called_once_with(file=Path(path), mode="w", encoding="utf-8")

        assert path == f"test/{Galadriel.name}.yml"
        mock_load.assert_not_called()

    def test_save_platform_with_yml_path(self, mock_open: MagicMock, mock_load: MagicMock):
        """Test the `save_platform` function with a path that ends in .yml."""
        path = save_platform(path="test/test.yml", platform=build_platform(Galadriel.runcard))

        mock_open.assert_called_once_with(file=Path(path), mode="w", encoding="utf-8")

        assert path == "test/test.yml"
        mock_load.assert_not_called()


class TestBuildPlatformCornerCases:
    """Unit tests for the corner cases of the `build_platform` function.."""

    def test_build_platform_passing_invalid_runcard_argument(self):
        """Test build method, with invalid argument."""
        for runcard in [None, 1, 1.0, True, False, [], ()]:
            with pytest.raises(ValueError) as wrong_runcard_error:
                ql.build_platform(runcard=runcard)
            (msg,) = wrong_runcard_error.value.args
            assert (
                msg
                == f"Incorrect type for `runcard` argument in `build_platform()`. Expected (str | dict), got: {type(runcard)}"
            )

    def test_build_method_with_new_drivers(self):
        """Test build method with the new drivers."""
        with pytest.raises(NotImplementedError, match="New drivers are not supported yet"):
            _ = ql.build_platform(runcard="_", new_drivers=True)

    def test_platform_serialization_from_imported_dict(self):
        """Test platform serialization by building a platform, saving it and then load it back again twice. Starting from a given dict."""
        original_dict = copy.deepcopy(Galadriel.runcard)
        # Check that the new serialization with ruamel.yaml.YAML().dump works for different formats...
        original_dict["digital"]["gates"]["Y(0)"][0]["pulse"]["phase"] = 1.6707963267948966  # Test long decimals
        original_dict["instruments"][0]["awg_sequencers"][0]["intermediate_frequency"] = 100_000_000  # Test underscores
        original_dict["instruments"][1]["awg_sequencers"][0]["sampling_rate"] = 7.24730e09  # Test scientific notation

        original_platform = ql.build_platform(original_dict)

        path = save_platform(path="./test.yml", platform=original_platform)
        saved_platform = ql.build_platform(path)

        new_path = save_platform(path="./test.yml", platform=saved_platform)
        new_saved_platform = ql.build_platform(new_path)

        original_platform_dict = original_platform.to_dict()
        saved_platform_dict = saved_platform.to_dict()
        new_saved_platform_dict = new_saved_platform.to_dict()

        assert (
            original_platform_dict
            == saved_platform_dict
            == new_saved_platform_dict
        )
        os.remove(path)  # Cleaning generated file

@patch("qililab.data_management.os.makedirs")
@patch("qililab.data_management.h5py.File")
class TestResultsData:
    """Unit tests for the `save_results` function."""

    def test_save_results_path(self, mock_file: MagicMock, mock_makedirs: MagicMock):
        """Test that the path created to save the results is correct."""

        path = save_results(results=np.array([]), loops={"one_loop": np.arange(0, 100, 0.1)}, data_path="", name="test")

        mock_file.assert_called_once_with(f"{path}/results.h5", "w")
        assert mock_makedirs.call_count == 2
        mock_makedirs.assert_called_with(path)

        assert path.endswith("_test")

    def test_load_results(self, mock_file: MagicMock, mock_makedirs: MagicMock):
        """Test that the path created to save the results is correct."""

        path = "this_is_a_test/results.h5"
        _ = load_results(path=path)

        mock_file.assert_called_once_with(path, "r")
        mock_makedirs.assert_not_called()



import os

import pytest

from qililab.qprogram.qblox_compiler import QbloxCompiler
from qililab.utils.serialization import (
    DeserializationError,
    SerializationError,
    deserialize,
    deserialize_from,
    serialize,
    serialize_to,
)
from qililab.waveforms import Gaussian


class TestSerialization:
    @pytest.mark.xfail(reason="Fails because YAML is instantiated as unsafe.")
    def test_serialization_of_not_registered_class_raises_error(self):
        compiler = QbloxCompiler()

        with pytest.raises(SerializationError):
            _ = serialize(compiler)

        with pytest.raises(SerializationError):
            serialize_to(compiler, "compiler.yml")

    def test_deserialization_with_wrong_yaml_raises_error(self):
        not_valid_yaml = "&id006 !QProgram"

        with pytest.raises(DeserializationError):
            _ = deserialize(not_valid_yaml)

        with open("not_valid_yaml.yml", "w", encoding="utf-8") as file:
            file.write(not_valid_yaml)
        with pytest.raises(DeserializationError):
            _ = deserialize_from("not_valid_yaml.yml")
        os.remove("not_valid_yaml.yml")

    def test_deserialization_with_wrong_cls_raises_error(self):
        serialized = "!Square {amplitude: 1.0, duration: 2000}\n"

        with pytest.raises(DeserializationError):
            _ = deserialize(serialized, Gaussian)

        with open("waveform.yml", "w", encoding="utf-8") as file:
            file.write(serialized)
        with pytest.raises(DeserializationError):
            _ = deserialize_from("waveform.yml", Gaussian)
        os.remove("waveform.yml")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class to interface with the local oscillator RohdeSchwarz SGS100A
"""

import warnings
from dataclasses import dataclass

from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue, RohdeSchwarzSGS100A


@InstrumentFactory.register
class SGS100A(Instrument):
    """Rohde & Schwarz SGS100A class

    Args:
        name (InstrumentName): name of the instrument
        device (RohdeSchwarz_SGS100A): Instance of the qcodes SGS100A class.
        settings (SGS100ASettings): Settings of the instrument.
    """

    name = InstrumentName.ROHDE_SCHWARZ

    @dataclass
    class SGS100ASettings(Instrument.InstrumentSettings):
        """Contains the settings of a specific signal generator.

        Args:
            power (float): Power of the instrument. Value range is (-120, 25).
            frequency (float): Frequency of the instrument. Value range is (1e6, 20e9).
        """

        power: float
        frequency: float
        rf_on: bool = True
        alc: bool = True
        iq_modulation: bool = False
        iq_wideband: bool = True

    settings: SGS100ASettings
    device: RohdeSchwarzSGS100A
    freq_top_limit: float
    freq_bot_limit: float
    device_initialized: bool = False

    @property
    def power(self):
        """SignalGenerator 'power' property.

        Returns:
            float: settings.power.
        """
        if self.device_initialized:
            if not -120 <= self.settings.power <= 25:
                raise ValueError(
                    f"Value set for power is outside of the allowed range [-120,25]: {self.settings.power}"
                )
        return self.settings.power

    @property
    def frequency(self):
        """SignalGenerator 'frequency' property.

        Returns:
            float: settings.frequency.
        """
        if self.device_initialized:
            if not self.freq_bot_limit <= self.settings.frequency <= self.freq_top_limit:
                raise ValueError(
                    f"Value set for frequency is outside of the allowed range [{self.freq_bot_limit}, {self.freq_top_limit}]: {self.settings.frequency}"
                )
        return self.settings.frequency

    @property
    def rf_on(self):
        """SignalGenerator 'rf_on' property.
        Returns:
            bool: settings.rf_on.
        """
        return self.settings.rf_on

    @property
    def alc(self):
        """SignalGenerator "Automatic Level Control" property.
        Returns:
            bool: settings.alc.
        """
        return self.settings.alc

    @property
    def iq_modulation(self):
        """SignalGenerator 'IQ state' property.
        Returns:
            bool: settings.iq_modulation.
        """
        return self.settings.iq_modulation

    @property
    def iq_wideband(self):
        """SignalGenerator "I/Q wideband" property.
        Returns:
            bool: settings.iq_wideband.
        """
        return self.settings.iq_wideband

    def to_dict(self):
        """Return a dict representation of the SignalGenerator class."""
        return dict(super().to_dict().items())

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set R&S dbm power and frequency. Value ranges are:
        - power: (-120, 25).
        - frequency (1e6, 20e9).
        """
        if parameter == Parameter.POWER:
            self.settings.power = float(value)
            if self.is_device_active():
                self.device.power(self.power)
            return
        if parameter == Parameter.LO_FREQUENCY:
            self.settings.frequency = float(value)
            if self.is_device_active():
                self.device.frequency(self.frequency)
            return
        if parameter == Parameter.RF_ON:
            self.settings.rf_on = bool(value)
            if self.is_device_active():
                if value:
                    self.turn_on()
                else:
                    self.turn_off()
            return
        if parameter == Parameter.ALC:
            self.settings.alc = bool(value)
            if self.is_device_active():
                status = "ONT"
                if not value:
                    status = "OFF"
                self.device.write(f":SOUR:POW:ALC:STAT {status}")
            return
        if parameter == Parameter.IQ_WIDEBAND:
            self.settings.iq_wideband = bool(value)
            if self.is_device_active():
                status = str(1)
                if not value:
                    status = str(0)
                self.device.write(f"SOUR:IQ:WBST {status}")
            return
        if parameter == Parameter.IQ_MODULATION:
            self.settings.iq_modulation = bool(value)
            if self.is_device_active():
                self.device.IQ_state(self.iq_modulation)
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None) -> ParameterValue:
        if parameter == Parameter.POWER:
            return self.settings.power
        if parameter == Parameter.LO_FREQUENCY:
            return self.settings.frequency
        if parameter == Parameter.RF_ON:
            return self.settings.rf_on
        if parameter == Parameter.ALC:
            return self.settings.alc
        if parameter == Parameter.IQ_WIDEBAND:
            return self.settings.iq_wideband
        if parameter == Parameter.IQ_MODULATION:
            return self.settings.iq_modulation
        raise ParameterNotFound(self, parameter)

    @check_device_initialized
    def get_rs_options(self):
        """Returns the options of the R&S
        Returns:
            str: The query returns a list of options. The options are returned at fixed positions in a comma-separated string. A zero is returned for options that are not installed.
        """
        return self.device.ask("*OPT?")

    @check_device_initialized
    def initial_setup(self):
        """performs an initial setup"""
        self.device_initialized = True
        device_mixer = self.get_rs_options().split(",")[-1]
        if device_mixer == "SGS-B112" or device_mixer == "SGS-B112V":
            self.freq_top_limit = 12.75e9
        elif device_mixer == "SGS-B106" or device_mixer == "SGS-B106V":
            self.freq_top_limit = 6e9
        if device_mixer == "SGS-B112" or device_mixer == "SGS-B106":
            self.freq_bot_limit = 1e6
        elif device_mixer == "SGS-B112V" or device_mixer == "SGS-B106V":
            self.freq_bot_limit = 80e6
        self.device.power(self.power)
        self.device.frequency(self.frequency)
        if self.alc:
            self.device.write(":SOUR:POW:ALC:STAT ONT")
        else:
            self.device.write(":SOUR:POW:ALC:STAT OFF")
        if device_mixer == "SGS-B112V" and self.iq_modulation:
            self.device.IQ_state(self.iq_modulation)
            if self.iq_wideband:
                if self.frequency < 2.5e9:
                    warnings.warn(
                        f"LO frequency below 2.5GHz only allows for IF sweeps of ±{self.frequency * 0.2:,.2e} Hz",
                        ResourceWarning,
                    )
                self.device.write("SOUR:IQ:WBST 1")
            else:
                if self.frequency < 1e9:
                    freq = self.frequency * 0.05
                    warnings.warn(
                        f"Deactivated wideband & LO frequency below 1GHz allows for IF sweeps of ±{freq:,.2e} Hz",
                        ResourceWarning,
                    )
                else:
                    warnings.warn("Deactivated wideband allows for IF sweeps of ±50e6 Hz", ResourceWarning)
                self.device.write("SOUR:IQ:WBST 0")
        elif not self.iq_modulation:
            self.device.IQ_state(self.iq_modulation)
        if self.rf_on:
            self.device.on()
        else:
            self.device.off()

    @check_device_initialized
    def turn_on(self):
        """Start generating microwaves."""
        if not self.settings.rf_on:
            self.device.off()  # this avoids the initialisation to overwrite the runcard
        else:
            self.device.on()

    @check_device_initialized
    def turn_off(self):
        """Stop generating microwaves."""
        self.settings.rf_on = False
        self.device.off()

    @check_device_initialized
    def reset(self):
        """Reset instrument."""



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import StringIO
from pathlib import Path
from typing import Any, TypeVar, overload

from qililab.yaml import yaml

T = TypeVar("T")


class SerializationError(Exception):
    """Custom exception for serialization errors."""


class DeserializationError(Exception):
    """Custom exception for deserialization errors."""


def serialize(obj: Any) -> str:
    """Serialize an object to a YAML string.

    Args:
        obj (Any): The object to serialize.

    Raises:
        SerializationError: If serialization fails.

    Returns:
        str: The serialized YAML string.
    """
    try:
        with StringIO() as stream:
            yaml.dump(obj, stream)
            return stream.getvalue()
    except Exception as e:
        raise SerializationError(f"Failed to serialize object {e}") from e


def serialize_to(obj: Any, file: str) -> None:
    """Serialize an object to a YAML file.

    Args:
        obj (Any): The object to serialize.
        file (str): The file path where the YAML data will be written.

    Raises:
        SerializationError: If serialization to file fails.
    """
    try:
        yaml.dump(obj, Path(file))
    except Exception as e:
        raise SerializationError(f"Failed to serialize object {e} to file {file}") from e


@overload
def deserialize(string: str) -> Any: ...


@overload
def deserialize(string: str, cls: type[T]) -> T: ...


def deserialize(string: str, cls: type[T] | None = None) -> Any | T:
    """Deserialize a YAML string to an object.

    Args:
        string (str): The YAML string to deserialize.
        cls (type[T], optional): The class type to cast the deserialized object to. Defaults to None.

    Raises:
        DeserializationError: If deserialization fails or the resulting object is not of the specified type.

    Returns:
        Any | T: The deserialized object, optionally cast to the specified class type.
    """
    try:
        with StringIO(string) as stream:
            result = yaml.load(stream)
    except Exception as e:
        raise DeserializationError(f"Failed to deserialize YAML string: {e}") from e
    if cls is not None and not isinstance(result, cls):
        raise DeserializationError(f"Deserialized object is not of type {cls.__name__}")
    return result


@overload
def deserialize_from(file: str) -> Any: ...


@overload
def deserialize_from(file: str, cls: type[T]) -> T: ...


def deserialize_from(file: str, cls: type[T] | None = None) -> Any | T:
    """Deserialize a YAML file to an object.

    Args:
        file (str): The file path of the YAML file to deserialize.
        cls (type[T], optional): The class type to cast the deserialized object to. Defaults to None.

    Raises:
        DeserializationError: If deserialization fails or the resulting object is not of the specified type.

    Returns:
        Any | T: The deserialized object, optionally cast to the specified class type.
    """
    try:
        result = yaml.load(Path(file))
    except Exception as e:
        raise DeserializationError(f"Failed to deserialize YAML string {e} from file {file}") from e
    if cls is not None and not isinstance(result, cls):
        raise DeserializationError(f"Deserialized object is not of type {cls.__name__}")
    return result



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from qililab.qprogram.operations.operation import Operation
from qililab.yaml import yaml


@yaml.register_class
class SetFrequency(Operation):
    def __init__(self, bus: str, frequency: float) -> None:
        super().__init__()
        self.bus: str = bus
        self.frequency: float = frequency



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Top level module from which all Qililab basic functions and classes can be directly imported."""

# isort: skip_file
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("qililab")
except PackageNotFoundError:
    __version__ = "0.0.0"

import contextlib

from .about import about
from .config import logger
from .data_management import build_platform, load_results, save_platform, save_results
from .execute_circuit import execute
from .qprogram import Calibration, CrosstalkMatrix, Domain, QbloxCompiler, QProgram, QuantumMachinesCompiler, Experiment
from .result import ExperimentResults, stream_results
from .typings import Parameter
from .utils.serialization import serialize, serialize_to, deserialize, deserialize_from
from .waveforms import IQPair, Square, Gaussian, FlatTop, Arbitrary, DragCorrection, Waveform, Ramp, Chained

# moving circuit_transpiler module imports here because it has instruments module dependencies so circular imports can be avoided
from .digital import Drag, Wait
from .analog import AnnealingProgram  # same as circuit transpiler, top modules should be imported at top


__all__ = [
    "AnnealingProgram",
    "Arbitrary",
    "Calibration",
    "Chained",
    "CrosstalkMatrix",
    "Domain",
    "Drag",
    "DragCorrection",
    "Experiment",
    "ExperimentResults",
    "FlatTop",
    "Gaussian",
    "IQPair",
    "Parameter",
    "QProgram",
    "QbloxCompiler",
    "QuantumMachinesCompiler",
    "Ramp",
    "Square",
    "Wait",
    "Waveform",
    "__version__",
    "about",
    "build_platform",
    "deserialize",
    "deserialize_from",
    "execute",
    "load_results",
    "logger",
    "save_platform",
    "save_results",
    "serialize",
    "serialize_to",
    "stream_results",
]


with contextlib.suppress(NameError, ImportError):
    # Since Ipython magic methods can only be imported from inside a Jupyter Notebook,
    # here we first check that `get_ipython` exists (which means we are inside a Jupyter Notebook)
    get_ipython()  # type: ignore  # noqa: F821
    from .slurm import submit_job as submit_job  # pragma: no cover
